# -*- coding: utf-8 -*-
#
# Copyright (C) 2011, 2012, 2013, 2015 OUI Technology Ltd.
# Copyright (C) 2019, 2020 Tomáš Cerha <t.cerha@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import lcg
import wiking
from wiking.cms import CMSExtensionModule, Roles
import pytis.data as pd
from pytis.presentation import Editable, Field, Action, SelectionType, computer
import pytis.web as pw

_ = lcg.TranslatableTextFactory('wiking-cms')


class CryptoNames(CMSExtensionModule):
    """Management of encryption passwords.

    This module just lists defined encryption areas.  The encryption areas are
    defined in the database by the application developer and can't be changed
    here.  The actual password managament is performed by CryptoKeys
    side-form module.

    """
    class Spec(wiking.Specification):
        table = 'cms_crypto_names'
        title = _("Crypto Areas")

        def fields(self):
            return (
                Field('name', _("Name")),
                Field('description', _("Description")),
            )
        sorting = (('name', pd.ASCENDENT,),
                   )
        bindings = (wiking.Binding('keys', _("Users and Keys"), 'CryptoKeys', 'name',
                                   form=pw.ItemizedView),
                    )

    def _authorized(self, req, action, **kwargs):
        if action in ('view', 'list'):
            return req.check_roles(Roles.CRYPTO_ADMIN)
        else:
            return False


class CryptoKeys(CMSExtensionModule):
    """Management of keys and users.

    It uses a special set of actions that resemble standard editing actions but
    are different from them:

    - Insert new key: This creates initial key for an encryption area.  This
      can be done only once for each of the areas.  The inserted keys can be
      given to other users using copy action.

    - Copy key: Copy an existing key to another user.

    - Change password: Change password of the key for the given user.

    - Delete key: Remove a user from the given key.  This can be only done if
      another copy of the key exists.

    All the actions are performed using special database functions.

    """
    class Spec(wiking.Specification):
        table = 'cms_crypto_keys'
        title = _("Users and Encryption Keys")

        def fields(self):
            return (
                Field('key_id', _("Id"), editable=Editable.NEVER),
                Field('name', _("Name"), not_null=True,
                      codebook='CryptoNames', editable=Editable.NEVER),
                Field('uid', _("User"), not_null=True, codebook='Users',
                      selection_type=SelectionType.CHOICE, editable=Editable.ONCE),
                Field('new_uid', _("New user"), not_null=True, codebook='Users',
                      selection_type=SelectionType.CHOICE, type=pd.Integer,
                      virtual=True, runtime_filter=computer(self._new_uid_filter)),
                Field('key', _("Key")),
                Field('remove', _("Action"), virtual=True,
                      computer=computer(lambda r: _("Remove"))),
                Field('old_password', _("Current password"),
                      type=pd.Password, verify=False, virtual=True),
                Field('new_password', _("New password"),
                      type=pd.Password, virtual=True),
                Field('delete', virtual=True, computer=computer(lambda row: _("Remove"))),
            )

        def _new_uid_filter(self, row, name):
            assigned_users = wiking.module(self._resolver).assigned_users(row['name'])
            return pd.AND(*[pd.NE('uid', u) for u in assigned_users])
        sorting = (('uid', pd.ASCENDENT,),
                   ('name', pd.ASCENDENT,),
                   )
        columns = ('uid', 'name', 'delete',)
        actions = (
            Action('password', _("Change password"), descr=_("Change key password")),
            Action('adduser', _("Copy to user"), descr=_("Add another user of the key")),
        )

    _DB_FUNCTIONS = dict(CMSExtensionModule._DB_FUNCTIONS,
                         cms_crypto_delete_key=(('name_', pd.String(),),
                                                ('uid_', pd.Integer(),),
                                                ('force', pd.Boolean(),),),
                         cms_crypto_insert_key=(('name_', pd.String(),),
                                                ('uid_', pd.Integer(),),
                                                ('key_', pd.String(),),
                                                ('psw', pd.String(),),),
                         cms_crypto_copy_key=(('name_', pd.String(),),
                                              ('from_uid', pd.Integer(),),
                                              ('to_uid', pd.Integer(),),
                                              ('from_psw', pd.String(),),
                                              ('to_psw', pd.String(),),),
                         cms_crypto_change_password=(('id_', pd.Integer(),),
                                                     ('old_psw', pd.String(),),
                                                     ('new_psw', pd.String(),),),
                         )

    _TITLE_COLUMN = 'uid'
    _INSERT_LABEL = _("Create key")

    def _authorized(self, req, action, record=None, **kwargs):
        if action == 'view':
            return req.check_roles(Roles.CRYPTO_ADMIN) or self._check_uid(req, record, 'uid')
        elif action == 'password':
            return self._check_uid(req, record, 'uid')
        elif action in ('list', 'insert', 'delete', 'adduser'):
            return req.check_roles(Roles.CRYPTO_ADMIN)
        else:
            return False

    def _prefill(self, req):
        return dict(super(CryptoKeys, self)._prefill(req),
                    uid=req.user().uid())

    def _layout(self, req, action, record=None):
        if action == 'insert':
            layout = ('name', 'uid', 'new_password',)
        elif action == 'adduser':
            layout = ('key_id', 'name', 'uid', 'new_uid', 'old_password', 'new_password',)
        elif action == 'password':
            layout = ('key_id', 'name', 'uid', 'old_password', 'new_password',)
        else:
            layout = ('name', 'uid',)
        return layout

    def _columns(self, req):
        columns = super(CryptoKeys, self)._columns(req)
        if not req.has_param('_crypto_name'):
            columns = [c for c in columns if c != 'delete']
        return columns

    def _link_provider(self, req, uri, record, cid, **kwargs):
        if cid == 'delete':
            return req.make_uri(uri, key_id=record['key_id'].value(), action='delete')
        else:
            return super(CryptoKeys, self)._link_provider(req, uri, record, cid, **kwargs)

    def _list_form_kwargs(self, req, form_cls):
        kwargs = super(CryptoKeys, self)._list_form_kwargs(req, form_cls)
        if issubclass(form_cls, pw.ItemizedView) and req.check_roles(Roles.USER_ADMIN):
            kwargs['template'] = lcg.TranslatableText("%%(%s)s [%%(delete)s]" % self._TITLE_COLUMN)
        return kwargs

    def related(self, req, binding, record, uri):
        if 'name' in record:
            req.set_param('_crypto_name', record['name'])
        return super(CryptoKeys, self).related(req, binding, record, uri)

    def _actions(self, req, record):
        actions = super(CryptoKeys, self)._actions(req, record)
        if record is None and req.has_param('_crypto_name'):
            condition = pd.EQ('name', req.param('_crypto_name'))
            try:
                count = self._data.select(condition)
            finally:
                try:
                    self._data.close()
                except Exception:
                    pass
            if count > 0:
                actions = [a for a in actions if a.id() != 'insert']
        return actions

    def _insert(self, req, record, transaction):
        key = wiking.generate_random_string(256)
        if not self._call_db_function('cms_crypto_insert_key',
                                      record['name'].value(),
                                      record['uid'].value(),
                                      key,
                                      record['new_password'].value(),
                                      transaction=transaction):
            raise pd.DBException(_("New key not created. Maybe it already exists?"))

    def _update(self, req, record, transaction):
        action = req.param('action')
        if action == 'password':
            if not self._call_db_function('cms_crypto_change_password',
                                          record['key_id'].value(),
                                          record['old_password'].value(),
                                          record['new_password'].value(),
                                          transaction=transaction):
                raise pd.DBException(_("Password not changed. Maybe invalid old password?"))
        elif action == 'adduser':
            if not self._call_db_function('cms_crypto_copy_key',
                                          record['name'].value(),
                                          record['uid'].value(),
                                          record['new_uid'].value(),
                                          record['old_password'].value(),
                                          record['new_password'].value(),
                                          transaction=transaction):
                raise pd.DBException(_("User not added. Maybe invalid old password?"))
        else:
            raise Exception('Unexpected action', action)

    def _delete(self, req, record, transaction):
        if not self._call_db_function('cms_crypto_delete_key',
                                      record['name'].value(),
                                      record['uid'].value(),
                                      False,
                                      transaction=transaction):
            raise pd.DBException(_("The user couldn't be deleted. "
                                   "Maybe he is the last key holder?"))

    def action_adduser(self, req, record, action='adduser'):
        return super(CryptoKeys, self).action_update(req, record, action=action)

    def action_password(self, req, record=None):
        return self.action_update(req, record=record, action='password')

    def assigned_users(self, name):
        # Internal method for Spec class, don't use it elsewhere
        return self._data.select_map(lambda row: row['uid'], condition=pd.EQ('name', name))

    def assigned_names(self, uid):
        """Return sequence of crypto names assigned to user identified by uid.

        Arguments:

          uid -- user uid, integer

        """
        return self._data.select_map(lambda row: row['name'], condition=pd.EQ('uid', pd.ival(uid)))

    def clear_crypto_passwords(self, req, user):
        # Just a hack to allow clearing passwords on logout
        req.set_cookie(self._CRYPTO_COOKIE, None, secure=True)
        self._call_db_function('cms_crypto_lock_passwords', user.uid())



# -*- coding: utf-8 -*-
#
# Copyright (C) 2006-2018 OUI Technology Ltd.
# Copyright (C) 2019-2020 Tomáš Cerha <t.cerha@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re

import lcg
import pytis
import wiking

_ = lcg.TranslatableTextFactory('wiking')
Part = lcg.Html5Exporter.Part


class MinimalExporter(lcg.Html5Exporter):
    _PAGE_STRUCTURE = (
        Part('main'),
        Part('bottom-bar'),
    )

    def _head(self, context):
        g = context.generator()
        try:
            uri = context.req().module_uri('Resources')
        except Exception:
            uri = '_resources'
        return super(MinimalExporter, self)._head(context) + \
            [g.link(rel='stylesheet', type='text/css', href='/%s/default.css' % uri)]

    def _meta(self, context):
        import wiking
        return (('generator', 'Wiking %s, LCG %s, Pytis %s' %
                 (wiking.__version__, lcg.__version__, pytis.__version__)),)

    def _main(self, context):
        g = context.generator()
        return (g.h1(context.node().title()),
                super(MinimalExporter, self)._content(context))

    def _bottom_bar(self, context):
        g = context.generator()
        import wiking
        return (g.hr(),
                g.span(g.a("Wiking", href="http://www.freebsoft.org/wiking") + ' ' +
                       wiking.__version__))


class Exporter(lcg.StyledHtmlExporter, lcg.Html5Exporter):

    class Context(lcg.Html5Exporter.Context):

        def _init_kwargs(self, req=None, layout='default', **kwargs):
            super(Exporter.Context, self)._init_kwargs(timezone=req.timezone(), **kwargs)
            assert layout in pytis.util.public_attr_values(self._exporter.Layout), layout
            self._req = req
            self._layout = layout
            node = top_node = self.node()
            root = node.root()
            while top_node and top_node.parent() and top_node.parent() is not root:
                top_node = top_node.parent()
            self._top_node = top_node
            application = wiking.module.Application
            self._panels = application.panels(req, self.lang())
            # Allow simple access to some often used data through context attributes ...
            # These attributes are not the part of the official context extension (such as the
            # 'req()' method, so their use should be limited to this module only!
            self.has_submenu = (wiking.cfg.show_submenu and
                                any(not n.hidden() for n in top_node.children()))
            self.application = application
            # Make sure that Prototype.js is always loaded first, so that it is
            # available in any other scripts.
            scripts = ('prototype.js', 'effects.js', 'gettext.js', 'lcg.js', 'wiking.js')
            for filename in scripts + tuple(wiking.cfg.extra_scripts):
                self.resource(filename)
            if self.lang() != 'en':
                self.resource('wiking.%s.po' % self.lang())  # Translations for Javascript

        def req(self):
            """Return the current request as a 'wiking.Request' instance.

            This method is the official Wiking extension of LCG export context.

            """
            return self._req

        def panels(self):
            """Return the list of 'wiking.Panel' instances to be displayed on the page."""
            return self._panels

        def top_node(self):
            """Return the top level parent of this node before the root node."""
            return self._top_node

        def layout(self):
            """Return the current export layout as one of 'Exporter.Layout' constants."""
            return self._layout

    class Layout:
        """Enumeration of output document layout styles."""
        DEFAULT = 'default'
        """Default Wiking layout wrapping the page content in menus, panels etc."""
        FRAME = 'frame'
        """Frame layout displaying just the document content wrapped in html body.

        This layout is typically useful for rendering the IFRAME content.  The
        exported document content is wrapped into HTML body with HTML head
        automatically created.

        """

    class MainMenu(lcg.FoldableTree):

        def _javascript_widget_class(self, context):
            return 'wiking.MainMenu'

        def _css_class_name(self, context):
            return 'foldable-tree-widget'

    _PAGE_STRUCTURE = (
        Part('root', content=(
            Part('root-wrap', content=(
                Part('top', aria_label=_("Page heading"), role='banner', content=(
                    Part('top-wrap', content=(
                        Part('top-bar', content=(
                            Part('top-content'),
                            Part('top-controls'),
                        )),
                        Part('menu',
                             aria_label=_("Main navigation"),
                             accesskey="3",
                             role='navigation'),
                    )),
                )),
                Part('main', content=(
                    Part('main-wrap', content=(
                        # Part('breadcrumbs'),
                        Part('submenu', role='navigation', aria_hidden='true'),
                        Part('page',
                             aria_label=_("Main content"),
                             aria_labelledby='main-heading',
                             role='main',
                             content=(
                                 Part('heading'),
                                 Part('messages'),
                                 Part('content'),
                             )),
                        Part('panels'),
                    )),
                )),
                Part('bottom', content=(
                    Part('bottom-wrap', content=(
                        Part('bottom-bar'),
                        Part('footer', aria_label=_("Page footer"), role='contentinfo'),
                    )),
                )),
            )),
        )),
    )
    _UNSAFE_CHARS = re.compile(r"[^a-zA-Z0-9_-]")

    def _safe_css_id(self, id):
        return self._UNSAFE_CHARS.sub('-', id)

    def _uri_node(self, context, node, lang=None):
        uri = node.id()
        if not uri.startswith('/'):
            uri = '/' + uri
        if '#' in uri:
            uri, anchor = uri.split('#', 1)
            args = (anchor,)
        else:
            args = ()
        return context.generator().uri(uri, *args, setlang=lang)

    def _resource_uri_prefix(self, context, resource):
        return context.req().module_uri('Resources')

    def _uri_resource(self, context, resource):
        uri = super(Exporter, self)._uri_resource(context, resource)
        # Minor hack to make CMS theme preview work (pass theme_id to styles heets).
        theme_id = context.req().param('preview_theme')
        if theme_id and isinstance(resource, lcg.Stylesheet):
            uri += '?preview_theme=%s' % theme_id
        return uri

    def _meta(self, context):
        import wiking
        result = [('generator', 'Wiking %s, LCG %s, Pytis %s' %
                   (wiking.__version__, lcg.__version__, pytis.__version__))]
        if wiking.cfg.viewport:
            result.append(('viewport', wiking.cfg.viewport))
        return result

    def _head(self, context):
        g = self._generator
        tags = super(Exporter, self)._head(context) + \
            [g.link(rel='alternate', type='application/rss+xml', title=p.title(), href=p.channel())
             for p in context.panels() if p.channel() is not None]
        if wiking.cfg.site_icon:
            tags.append(g.link(rel='shortcut icon', href='/favicon.ico'))
        req = context.req()
        server_uri = req.server_uri()
        return tags + [g.meta(property='og:' + name, content=val) for name, val in (
            ('title', context.node().title()),
            ('type', 'article'),
            ('url', server_uri + req.uri()),
            ('site_name', context.application.site_title(req)),
            ('description', context.application.site_subtitle(req)),
            ('image', server_uri + wiking.cfg.site_image if wiking.cfg.site_image else None),
        ) if val is not None]

    def _body_content(self, context):
        if context.layout() == self.Layout.FRAME:
            messages = self._messages(context)
            return (messages and self._generator.div(messages, id='messages') or '',
                    self._content(context))
        else:
            return super(Exporter, self)._body_content(context)

    def _body_attr(self, context):
        attr = super(Exporter, self)._body_attr(context)
        cls = ['page-id-' + self._safe_css_id(context.node().id().strip('/'))]
        cls.extend(['parent-page-id-' + self._safe_css_id(node.id().strip('/'))
                    for node in context.node().path()[1:-1]])
        cls.extend(('lang-%s' % context.lang(),
                    context.layout() + '-layout'))
        req = context.req()
        if req.maximized():
            cls.append('maximized')
        else:
            cls.append('non-maximized')
        if context.panels():
            cls.append('with-panels')
        if context.has_submenu:
            cls.append('with-submenu')
        cls.extend(context.application.body_class_names(req))
        return dict(attr,
                    onload=context.generator().js_call('new wiking.Handler'),
                    cls=' '.join(cls))

    def _site_title(self, context):
        g = self._generator
        title = context.application.site_title(context.req())
        subtitle = context.application.site_subtitle(context.req())
        content = g.strong(title, cls='title')
        if subtitle:
            content += (g.strong(g.noescape(' &ndash; '), cls='separator') +
                        g.strong(subtitle, cls='subtitle'))
        return content

    def _title(self, context):
        return context.node().title() + ' - ' + context.application.site_title(context.req())

    def _top_content(self, context):
        content = context.application.top_content(context.req())
        if content:
            return lcg.coerce(content).export(context)
        else:
            return None

    def _top_controls(self, context):
        return lcg.coerce(context.application.top_controls(context.req())).export(context)

    def _breadcrumbs(self, context):
        links = [lcg.link(n).export(context) for n in context.node().path()[1:]]
        # Translators: A label followed by location information in webpage navigation
        return _("You are here:") + ' ' + lcg.concat(links, separator=' / ')

    def _menu(self, context):
        g = self._generator
        return (g.div(g.span('', cls='menu-icon', tabindex=0),
                      aria_label=_("Menu"), aria_role='button', cls='menu-button'),
                self.MainMenu(context.node().root()).export(context))

    def _submenu(self, context):
        if not context.has_submenu:
            return None
        g = self._generator
        top = context.top_node()
        tree = lcg.FoldableTree(top, label=_("Local navigation: %s", top.title()),
                                tooltip=_("Expand/collapse complete menu hierarchy"))
        content = tree.export(context)
        req = context.req()
        application = context.application
        title = application.menu_panel_title(req)
        if title:
            content = lcg.concat(g.h3(title, role='presentation', aria_hidden='true'), content)
        bottom_content = application.menu_panel_bottom_content(req)
        if bottom_content:
            content = lcg.concat(content, bottom_content.export(context))
        return g.div(content, cls='menu-panel')

    def _messages(self, context):
        messages = context.req().messages()
        if messages:
            return [wiking.Message(message, kind=kind, formatted=formatted).export(context)
                    for message, kind, formatted in messages]
        else:
            return None

    def _heading(self, context):
        g = self._generator
        return g.h1(g.a(context.node().heading().export(context), tabindex=0,
                        name='main-heading', id='main-heading'))

    def _panels(self, context):
        if not context.panels():
            return None
        g = self._generator
        req = context.req()
        panels = context.panels()
        if not panels:
            return None
        result = []
        for panel in panels:
            title = g.a(panel.title(), name='panel-' + panel.id() + '-anchor', tabindex=0,
                        cls='panel-anchor')
            titlebar_content = panel.titlebar_content()
            if titlebar_content:
                title += titlebar_content.export(context)
            channel = panel.channel()
            if channel:
                # Translators: ``RSS channel'' is terminology idiom, see Wikipedia.
                # The placeholder %s is replaced by channel title.
                channel_title = _("RSS channel %s") + ' ' + panel.title()
                title += g.a(g.span('', cls='feed-icon'),
                             href=channel, aria_label=channel_title, title=channel_title,
                             type='application/rss+xml', cls='feed-link')
            content = panel.content()
            # Add a fake container to force the heading level start at 4.
            lcg.Container(lcg.Section('', lcg.Section('', content)))
            result.append(g.div((g.h3(title),
                                 g.div(content.export(context), cls='panel-content')),
                                id='panel-' + self._safe_css_id(panel.id()), cls='panel',
                                role='complementary',
                                aria_label=panel.accessible_title()))
        extra_content = context.application.right_panels_bottom_content(req)
        if extra_content:
            result.append(g.div(extra_content.export(context), cls='panels-bottom-content'))
        return result

    def _last_change(self, context):
        # Currently unused, left here just to have the translation.
        # Translators: Information about last change of a webpage (when and who)
        return _("Last change: %(date)s, %(user)s")

    def _bottom_bar(self, context):
        req = context.req()
        left = context.application.bottom_bar_left_content(req)
        right = context.application.bottom_bar_right_content(req)
        if left or right:
            g = self._generator
            result = [g.hr()]
            if left:
                content = lcg.coerce(left)
                if right:
                    result.append(g.span(content.export(context), cls="left"))
                else:
                    result.append(g.div(content.export(context), cls="left"))
            if right:
                content = lcg.coerce(right)
                result.append(g.span(content.export(context), cls="right"))
            return result
        else:
            return None

    def _footer(self, context):
        content = context.application.footer_content(context.req())
        if content:
            return lcg.coerce(content).export(context)
        else:
            return None

    def uri(self, context, target, **kwargs):
        return (context.req().root() or '') + super().uri(context, target, **kwargs)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2010-2016 OUI Technology Ltd.
# Copyright (C) 2019-2021, 2025 Tomáš Cerha <t.cerha@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import cgi
import os
import wsgiref.util
import wsgiref.headers
import wiking

import http.client
import urllib.parse


class WsgiRequest(wiking.Request):
    """Wiking server interface implementation for WSGI.

    Implements the 'wiking.ServerInterface' API.

    """
    class Generator:
        pass

    def __init__(self, environ, start_response, encoding='utf-8'):
        """Initializes the request with WSGI parameters.

        environ, start_response -- same-named arguments to the WSGI application
            object callable, as per WSGI spec (PEP 333)

        """
        # It is important to avoid possible exceptions (such as UnicodeDecodeError)
        # here as they would escape Wiking's exception handling.  Thus further
        # sensitive processing should be done later on demand when instance methods
        # are called.
        self._environ = environ
        self._start_response = start_response
        self._root = self._environ.get('SCRIPT_NAME')
        self._uri = self._environ['PATH_INFO']
        self._params = {}
        if self.method() == 'OPTIONS':
            self._raw_params = {}
        else:
            self._raw_params = cgi.FieldStorage(fp=environ['wsgi.input'], environ=environ,
                                                keep_blank_values=True)
        self._unset_params = []
        self._response_headers_storage = []
        self._response_headers = wsgiref.headers.Headers(self._response_headers_storage)
        super(WsgiRequest, self).__init__(encoding=encoding)

    def root(self):
        return self._root

    def uri(self):
        return self._uri

    def unparsed_uri(self):
        uri = wsgiref.util.request_uri(self._environ)
        if uri:
            uri = urllib.parse.urlunsplit(('', '',) + urllib.parse.urlsplit(uri)[2:])
        return uri

    def method(self):
        return self._environ['REQUEST_METHOD']

    def param(self, name, default=None):
        def param_value(value):
            # Value processing not done in constructor (see the constructor comment).
            if isinstance(value, (tuple, list)):
                return tuple(param_value(v) for v in value)
            elif value.filename == '' and value.value == b'':
                # Empty file upload fields give this strange combination...
                return None
            elif value.filename:
                return wiking.FileUpload(value, self._encoding)
            else:
                return value.value
        try:
            return self._params[name]
        except KeyError:
            if name in self._unset_params:
                return default
            try:
                raw_value = self._raw_params[name]
            except KeyError:
                return default
            self._params[name] = value = param_value(raw_value)
            return value

    def params(self):
        return tuple(set(self._raw_params).union(set(self._params)) - set(self._unset_params))

    def set_param(self, name, value):
        if value is None:
            if name in self._params:
                del self._params[name]
            if name not in self._unset_params:
                self._unset_params.append(name)
        else:
            if name in self._unset_params:
                self._unset_params.remove(name)
            self._params[name] = value

    def header(self, name, default=None):
        if name == 'Content-Type':
            environ_name = 'CONTENT_TYPE'
        elif name == 'Content-Length':
            environ_name = 'CONTENT_LENGTH'
        else:
            environ_name = 'HTTP_' + name.replace('-', '_').upper()
        return self._environ.get(environ_name, default)

    def set_header(self, name, value, **params):
        def encode(string):
            # Hack: Non latin-1 characters will lead to
            # "TypeError: http header must be encodable in latin1"
            # when start_http_response() is called.  There doesn't seem to be
            # any official documentation how this should be solved correctly,
            # but this hack seems to work.
            return string.encode(self._encoding).decode('latin1')
        # See wsgiref.headers.Headers.add_header for the documentation.
        self._response_headers.add_header(name, encode(value), **{k: encode(v) for k, v in params.items()})

    def port(self):
        port = self._environ['SERVER_PORT']
        if not port:
            return None
        return int(port)

    def https(self):
        return self._environ['wsgi.url_scheme'] == 'https'

    def remote_host(self):
        return self._environ.get('REMOTE_HOST', self._environ.get('REMOTE_ADDR'))

    def server_hostname(self):
        return self.header('Host') or self._environ['SERVER_NAME']

    def primary_server_hostname(self):
        # PEP http://www.python.org/dev/peps/pep-3333/ is not very clear on
        # what SERVER_NAME should exactly contain and we don't know of any
        # other method how to retrieve the unique server name (as descrined
        # in the docstring of this method in the parent class) under WSGI.
        # So at least under Apache/mod_wsgi, it is necessary to set the
        # environment variable 'wiking.server_hostname' according the
        # ServerName directive whenever there is one or more ServerAlias
        # directives because SERVER_NAME doesn't contain what we need.
        return None

    def start_http_response(self, status_code):
        self._start_response(
            '%d %s' % (status_code, http.client.responses[status_code]),
            self._response_headers_storage
        )

    def option(self, name, default=None):
        return self._environ.get('wiking.' + name, default)


class WsgiEntryPoint:
    """WSGI entry point.

    This class implements a WSGI specific wrapper of 'wiking.Handler'.

    An instance of this class is created below to serve as mod_wsgi entry
    point.  The instance is callable and will be called to serve the request.

    """

    def __init__(self):
        self._handler = None

    def __call__(self, environ, start_response):
        req = WsgiRequest(environ, start_response)
        handler = self._handler
        if handler is None:
            # Initialization is postponed until the first request, since we
            # need information from the environment to initialize the the
            # handler instance.
            handler = self._handler = wiking.Handler(req)
        return handler.handle(req)


application = WsgiEntryPoint()

# Hack to allow wsgi shell access, any better solution is welcome.
_wsgi_shell_config_file = os.path.expanduser('~/ispyd.ini')
if os.access(_wsgi_shell_config_file, os.R_OK):
    from ispyd.manager import ShellManager
    shell_manager = ShellManager(_wsgi_shell_config_file)
    from ispyd.plugins.wsgi import WSGIApplicationWrapper
    application = WSGIApplicationWrapper(application)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2012, 2013, 2015, 2016 OUI Technology Ltd.
# Copyright (C) 2019 Tomáš Cerha <t.cerha@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Update and compile source code in GIT repositories found in given directory.

This script simplifies updates of typical Wiking web application deployments
which run from git repositories specific for a particular web server's virtual
host.

All sudbirectories, which are git repositories are updated from git by running
"git pull" and if they contain a makefile, "make" is run as well.  This usually
compiles Python byte code and generates Gettext translations in repositories
typically needed for Wiking web applications (LCG, Pytis, Wiking etc).  Note
that you may need to run upgrade-db.py after this script.

"""
import os
import sys
import subprocess


def usage(message=None):
    text = ["Usage: %s directory" % os.path.basename(sys.argv[0])]
    if message:
        text.append(message)
        text.append("Use --help for more information.")
    else:
        text.append(__doc__.strip())
    sys.stderr.write('\n\n'.join(text) + '\n')
    sys.exit(1)


def call(command, cwd):
    result = subprocess.call(command.split(' '), cwd=cwd)
    if result != 0:
        sys.exit(1)


def update_src(directory):
    if not os.path.isdir(directory):
        usage("Not a directory: %s" % directory)
    repos = [d for d in os.listdir(directory)
             if os.path.isdir(os.path.join(directory, d, '.git'))]
    if not repos:
        usage("No subdirectories found: %s" % os.path.abspath(directory))
    for repo in repos:
        print("Updating repository: %s" % repo)
        subdir = os.path.join(directory, repo)
        call('git pull', subdir)
        if ((os.path.exists(os.path.join(subdir, 'Makefile')) or
             os.path.exists(os.path.join(subdir, 'makefile')))):
            call('make', subdir)

if __name__ == '__main__':
    if '--help' in sys.argv:
        usage()
    if len(sys.argv) != 2:
        usage("Invalid number of arguments.")
    update_src(sys.argv[1])



# -*- coding: utf-8 -*-
#
# Copyright (C) 2006-2016 OUI Technology Ltd.
# Copyright (C) 2019-2025 Tomáš Cerha <t.cerha@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import glob
import sqlalchemy

import pytis.data.gensqlalchemy as sql
import pytis.data as pd
from pytis.data.dbdefs import and_, or_, coalesce, func, ival, null, select, stype, sval
from .wiking_db import Base_CachingTable, CommonAccesRights

current_timestamp_0 = sqlalchemy.sql.functions.Function('current_timestamp', ival(0))

name_is_not_null = sql.SQLFlexibleValue('name_not_null', default=True)

upgrade_directory = os.path.abspath(os.path.join(os.path.realpath(os.path.dirname(__file__)),
                                                 '..', '..', '..', 'upgrade'))
upgrade_files = glob.glob(os.path.join(upgrade_directory, 'upgrade.*.sql'))
assert upgrade_files, 'No upgrade files found in upgrade directory: %s' % upgrade_directory
last_upgrade_script = max(upgrade_files)
version = int(last_upgrade_script[len(upgrade_directory) + len(os.sep) + 8:-4])


class SQLView(sql.SQLView):

    @classmethod
    def _reference(cls, kind):
        """Refer NEW and OLD columns in rules."""
        class Reference:
            def __init__(self, kind):
                self._kind = kind
            def __getattr__(self, name):
                return sqlalchemy.literal_column(self._kind + '.' + name)
        return Reference(kind)


class cms_database_version(sql.SQLTable):
    name = 'cms_database_version'
    fields = (
        sql.Column('version', pd.Integer()),
    )
    init_columns = ('version',)
    init_values = ((version,),)


class cms_languages(CommonAccesRights, Base_CachingTable):
    name = 'cms_languages'
    fields = (
        sql.PrimaryColumn('lang_id', pd.Serial()),
        sql.Column('lang', pd.String(minlen=2, maxlen=2, not_null=True),
                   unique=True),
    )
    init_columns = ('lang',)
    init_values = (('en',),)


class cms_config(CommonAccesRights, Base_CachingTable):
    name = 'cms_config'
    fields = (
        sql.PrimaryColumn('site', pd.String()),
        sql.Column('site_title', pd.String()),
        sql.Column('site_subtitle', pd.String()),
        sql.Column('allow_registration', pd.Boolean(not_null=True), default=True),
        sql.Column('login_is_email', pd.Boolean(not_null=True), default=False),
        sql.Column('registration_expiration', pd.Integer()),
        sql.Column('force_https_login', pd.Boolean(not_null=True), default=False),
        sql.Column('https_port', pd.Integer()),
        sql.Column('smtp_server', pd.String()),
        sql.Column('webmaster_address', pd.String()),
        sql.Column('bug_report_address', pd.String()),
        sql.Column('default_sender_address', pd.String()),
        sql.Column('upload_limit', pd.Integer()),
        sql.Column('session_expiration', pd.Integer()),
        sql.Column('default_language', pd.String(minlen=2, maxlen=2),
                   references=sql.a(sql.r.cms_languages.lang, onupdate='CASCADE')),
        sql.Column('theme_id', pd.Integer(),
                   references=sql.r.cms_themes),
    )
    init_columns = ('site',)
    init_values = (('*',),)


class cms_countries(CommonAccesRights, sql.SQLTable):
    name = 'cms_countries'
    fields = (
        sql.PrimaryColumn('country_id', pd.Serial(not_null=True)),
        sql.Column('country', pd.String(minlen=2, maxlen=2, not_null=True),
                   unique=True),
    )
    init_columns = ('country',)
    init_values = [(country,) for country in (
        'AD', 'AE', 'AF', 'AG', 'AI', 'AL', 'AM', 'AO', 'AQ', 'AR', 'AS', 'AT',
        'AU', 'AW', 'AX', 'AZ', 'BA', 'BB', 'BD', 'BE', 'BF', 'BG', 'BH', 'BI',
        'BJ', 'BL', 'BM', 'BN', 'BO', 'BQ', 'BR', 'BS', 'BT', 'BV', 'BW', 'BY',
        'BZ', 'CA', 'CC', 'CD', 'CF', 'CG', 'CH', 'CI', 'CK', 'CL', 'CM', 'CN',
        'CO', 'CR', 'CU', 'CV', 'CW', 'CX', 'CY', 'CZ', 'DE', 'DJ', 'DK', 'DM',
        'DO', 'DZ', 'EC', 'EE', 'EG', 'EH', 'ER', 'ES', 'ET', 'FI', 'FJ', 'FK',
        'FM', 'FO', 'FR', 'GA', 'GB', 'GD', 'GE', 'GF', 'GG', 'GH', 'GI', 'GL',
        'GM', 'GN', 'GP', 'GQ', 'GR', 'GS', 'GT', 'GU', 'GW', 'GY', 'HK', 'HM',
        'HN', 'HR', 'HT', 'HU', 'ID', 'IE', 'IL', 'IM', 'IN', 'IO', 'IQ', 'IR',
        'IS', 'IT', 'JE', 'JM', 'JO', 'JP', 'KE', 'KG', 'KH', 'KI', 'KM', 'KN',
        'KP', 'KR', 'KW', 'KY', 'KZ', 'LA', 'LB', 'LC', 'LI', 'LK', 'LR', 'LS',
        'LT', 'LU', 'LV', 'LY', 'MA', 'MC', 'MD', 'ME', 'MF', 'MG', 'MH', 'MK',
        'ML', 'MM', 'MN', 'MO', 'MP', 'MQ', 'MR', 'MS', 'MT', 'MU', 'MV', 'MW',
        'MX', 'MY', 'MZ', 'NA', 'NC', 'NE', 'NF', 'NG', 'NI', 'NL', 'NO', 'NP',
        'NR', 'NU', 'NZ', 'OM', 'PA', 'PE', 'PF', 'PG', 'PH', 'PK', 'PL', 'PM',
        'PN', 'PR', 'PS', 'PT', 'PW', 'PY', 'QA', 'RE', 'RO', 'RS', 'RU', 'RW',
        'SA', 'SB', 'SC', 'SD', 'SE', 'SG', 'SH', 'SI', 'SJ', 'SK', 'SL', 'SM',
        'SN', 'SO', 'SR', 'SS', 'ST', 'SV', 'SX', 'SY', 'SZ', 'TC', 'TD', 'TF',
        'TG', 'TH', 'TJ', 'TK', 'TL', 'TM', 'TN', 'TO', 'TR', 'TT', 'TV', 'TW',
        'TZ', 'UA', 'UG', 'UM', 'US', 'UY', 'UZ', 'VA', 'VC', 'VE', 'VG', 'VI',
        'VN', 'VU', 'WF', 'WS', 'YE', 'YT', 'ZA', 'ZM', 'ZW',
    )]


class roles(CommonAccesRights, Base_CachingTable):
    name = 'roles'
    fields = (
        sql.PrimaryColumn('role_id', pd.PgName()),
        sql.Column('name', pd.String()),
        sql.Column('system', pd.Boolean(not_null=True), default=False),
        sql.Column('auto', pd.Boolean(not_null=True), default=False),
    )
    init_columns = ('role_id', 'system', 'auto',)
    init_values = (
        ('anyone', 't', 't'),
        ('authenticated', 't', 't'),
        ('owner', 't', 't'),
        ('user', 't', 't'),
        ('registered', 't', 't'),
        ('cms-user-admin', 't', 'f'),
        ('cms-crypto-admin', 't', 'f'),
        ('cms-content-admin', 't', 'f'),
        ('cms-settings-admin', 't', 'f'),
        ('cms-mail-admin', 't', 'f'),
        ('cms-style-admin', 't', 'f'),
        ('cms-admin', 't', 'f'),
    )


class role_sets(CommonAccesRights, Base_CachingTable):
    name = 'role_sets'
    fields = (
        sql.PrimaryColumn('role_set_id', pd.Serial(not_null=True)),
        sql.Column('role_id', pd.PgName(not_null=True),
                   references=sql.a(sql.r.roles, onupdate='CASCADE', ondelete='CASCADE')),
        sql.Column('member_role_id', pd.PgName(not_null=True),
                   references=sql.a(sql.r.roles, onupdate='CASCADE', ondelete='CASCADE')),
    )
    unique = (('role_id', 'member_role_id',),)
    init_columns = ('role_id', 'member_role_id',)
    init_values = (
        ('cms-admin', 'cms-user-admin'),
        ('cms-admin', 'cms-crypto-admin'),
        ('cms-admin', 'cms-content-admin'),
        ('cms-admin', 'cms-settings-admin'),
        ('cms-admin', 'cms-mail-admin'),
        ('cms-admin', 'cms-style-admin'),
    )


class expanded_role(sql.SQLPlFunction):
    name = 'expanded_role'
    arguments = (sql.Column('role_id_', pd.PgName()),)
    result_type = pd.PgName()
    multirow = True
    stability = 'stable'


class unrelated_roles(sql.SQLFunction):
    name = 'unrelated_roles'
    arguments = (sql.Column('role_id', pd.PgName()),)
    result_type = roles
    multirow = True
    stability = 'stable'


class users(CommonAccesRights, Base_CachingTable):
    name = 'users'
    fields = (
        sql.PrimaryColumn('uid', pd.Serial(not_null=True)),
        sql.Column('login', pd.String(maxlen=64, not_null=True), unique=True),
        sql.Column('password', pd.String(not_null=True)),
        sql.Column('firstname',
                   pd.String(not_null=name_is_not_null.value(globals()))),
        sql.Column('surname',
                   pd.String(not_null=name_is_not_null.value(globals()))),
        sql.Column('nickname', pd.String()),
        sql.Column('user_', pd.String(not_null=True)),
        sql.Column('email', pd.String(not_null=True)),
        sql.Column('phone', pd.String()),
        sql.Column('address', pd.String()),
        sql.Column('uri', pd.String()),
        sql.Column('state', pd.String(not_null=True), default='new'),
        sql.Column('last_password_change', pd.DateTime(not_null=True)),
        sql.Column('since', pd.DateTime(not_null=True),
                   default=func.timezone(sval('GMT'), current_timestamp_0)),
        sql.Column('lang', pd.String(minlen=2, maxlen=2),
                   references=sql.a(sql.r.cms_languages.lang, onupdate='CASCADE',
                                    ondelete='SET NULL')),
        sql.Column('regexpire', pd.DateTime()),
        sql.Column('regcode', pd.String()),
        sql.Column('passexpire', pd.DateTime()),
        sql.Column('passcode', pd.String()),
        sql.Column('certauth', pd.Boolean(not_null=True), default=False),
        sql.Column('note', pd.String()),
        sql.Column('confirm', pd.Boolean(not_null=True), default=False),
        sql.Column('gender', pd.String(minlen=1, maxlen=1),
                   doc="[m]ale, [f]emale, NULL=unknown"),
    )
    access_rights = (('ALL', 'www-data',),)

    init_columns = ('login', 'password', 'firstname', 'surname', 'nickname', 'user_',
                    'email', 'state', 'last_password_change',)
    init_values = (('admin', 'plain:wiking', 'Wiking', 'Admin', 'Admin', 'Admin',
                    '-', 'enabled', '2012-01-01 00:00'),)


class cms_f_insert_or_update_user(sql.SQLPlFunction):
    name = 'cms_f_insert_or_update_user'
    arguments = (sql.Column('uid_', pd.Integer()),
                 sql.Column('login_', pd.String(maxlen=64)),
                 sql.Column('password_', pd.String()),
                 sql.Column('firstname_', pd.String()),
                 sql.Column('surname_', pd.String()),
                 sql.Column('nickname_', pd.String()),
                 sql.Column('user__', pd.String()),
                 sql.Column('email_', pd.String()),
                 sql.Column('phone_', pd.String()),
                 sql.Column('address_', pd.String()),
                 sql.Column('uri_', pd.String()),
                 sql.Column('state_', pd.String()),
                 sql.Column('last_password_change_', pd.DateTime()),
                 sql.Column('since_', pd.DateTime()),
                 sql.Column('lang_', pd.String(minlen=2, maxlen=2)),
                 sql.Column('regexpire_', pd.DateTime()),
                 sql.Column('regcode_', pd.String()),
                 sql.Column('passexpire_', pd.DateTime()),
                 sql.Column('passcode_', pd.String()),
                 sql.Column('certauth_', pd.Boolean()),
                 sql.Column('note_', pd.String()),
                 sql.Column('confirm_', pd.Boolean()),
                 sql.Column('gender_', pd.String(minlen=1, maxlen=1)),
                 )
    result_type = None


class role_members(CommonAccesRights, Base_CachingTable):
    name = 'role_members'
    fields = (
        sql.PrimaryColumn('role_member_id', pd.Serial(not_null=True)),
        sql.Column('role_id', pd.PgName(not_null=True),
                   references=sql.a(sql.r.roles, onupdate='CASCADE', ondelete='CASCADE')),
        sql.Column('uid', pd.Integer(not_null=True),
                   references=sql.a(sql.r.users, onupdate='CASCADE', ondelete='CASCADE')),
    )
    unique = (('role_id', 'uid',),)
    init_columns = ('role_id', 'uid')
    init_values = (('cms-admin', 1),)


class cms_v_role_members(CommonAccesRights, SQLView):
    name = 'cms_v_role_members'

    @classmethod
    def query(cls):
        m = sql.t.role_members.alias('m')
        r = sql.t.roles.alias('r')
        u = sql.t.users.alias('u')
        return select([m,
                       r.c.name.label('role_name'),
                       u.c.user_.label('user_name'),
                       u.c.login.label('user_login'),
                       ],
                      from_obj=[m.join(r, r.c.role_id == m.c.role_id).join(u, m.c.uid == u.c.uid)])
    insert_order = (role_members,)
    update_order = (role_members,)
    delete_order = (role_members,)
    no_insert_columns = ('role_member_id',)


class a_user_roles(CommonAccesRights, sql.SQLTable):
    name = 'a_user_roles'
    fields = (
        sql.Column('uid', pd.Integer(), index=True,
                   references=sql.a(sql.r.users, onupdate='CASCADE', ondelete='CASCADE')),
        sql.Column('role_id', pd.PgName(not_null=True),
                   references=sql.a(sql.r.roles, onupdate='CASCADE', ondelete='CASCADE')),
    )
    access_rights = (('ALL', 'www-data',),)


class update_user_roles(sql.SQLPlFunction, sql.SQLTrigger):
    name = 'update_user_roles'
    arguments = ()
    events = ()


class role_sets_update_user_roles_trigger(sql.SQLTrigger):
    name = 'role_sets_update_user_roles_trigger'
    table = role_sets
    events = ('insert', 'update', 'delete',)
    each_row = False
    body = update_user_roles


class role_members_update_user_roles_trigger(sql.SQLTrigger):
    name = 'role_members_update_user_roles_trigger'
    table = role_members
    events = ('insert', 'update', 'delete',)
    each_row = False
    body = update_user_roles


class cms_f_all_user_roles(sql.SQLFunction):
    """Return all user's roles and their included roles.
    Both explicitly assigned and implicit roles (such as 'anyone') are considered."""
    name = 'cms_f_all_user_roles'
    arguments = (sql.Column('uid_', pd.Integer()),)
    result_type = pd.PgName()
    multirow = True
    stability = 'stable'
    depends_on = (a_user_roles,)


class cms_f_role_member(sql.SQLPlFunction):
    name = 'cms_f_role_member'
    arguments = (sql.Column('uid_', pd.Integer()),
                 sql.Column('role_id_', pd.PgName()),)
    result_type = pd.Boolean()
    stability = 'stable'


class role_sets_cycle_check(sql.SQLFunction):
    name = 'role_sets_cycle_check'
    arguments = ()
    result_type = pd.Boolean()
    depends_on = (role_sets,)


class role_sets_trigger_after(sql.SQLPlFunction, sql.SQLTrigger):
    name = 'role_sets_trigger_after'
    arguments = ()
    table = role_sets
    position = 'after'
    events = ('insert', 'update', 'delete',)


class cms_sessions(CommonAccesRights, sql.SQLTable):
    name = 'cms_sessions'
    fields = (
        sql.PrimaryColumn('session_id', pd.Serial(not_null=True)),
        sql.Column('session_key', pd.String(not_null=True)),
        sql.Column('auth_type', pd.String(not_null=True)),
        sql.Column('uid', pd.Integer(not_null=True),
                   references=sql.a(sql.r.users, ondelete='CASCADE')),
        sql.Column('last_access', pd.DateTime()),
    )
    unique = (('uid', 'session_key',),)

class cms_session_history(CommonAccesRights, sql.SQLTable):
    name = 'cms_session_history'
    fields = (
        sql.PrimaryColumn('session_id', pd.Integer(not_null=True)),
        sql.Column('auth_type', pd.String(not_null=True)),
        sql.Column('uid', pd.Integer(not_null=True),
                   references=sql.a(sql.r.users, ondelete='CASCADE')),
        sql.Column('start_time', pd.DateTime(not_null=True)),
        sql.Column('end_time', pd.DateTime()),
    )

class cms_update_session_history(sql.SQLPlFunction, sql.SQLTrigger):
    name = 'cms_update_session_history'
    arguments = ()
    events = ()

class cms_sessions_trigger(sql.SQLTrigger):
    name = 'cms_sessions_trigger'
    table = cms_sessions
    events = ('insert', 'delete',)
    each_row = True
    body = cms_update_session_history

class cms_v_session_history(CommonAccesRights, SQLView):
    name = 'cms_v_session_history'
    @classmethod
    def query(cls):
        h = sql.t.cms_session_history.alias('h')
        u = sql.t.users.alias('u')
        return select(list(h.c) + [
            u.c.login,
            u.c.user_.label('user'),
            sqlalchemy.cast(h.c.end_time == null, sqlalchemy.Boolean()).label('active'),
            (coalesce(h.c.end_time, func.now()) - h.c.start_time).label('duration'),
        ], from_obj=[h.outerjoin(u, u.c.uid == h.c.uid)])


class cms_login_failures(CommonAccesRights, sql.SQLTable):
    name = 'cms_login_failures'
    fields = (
        sql.PrimaryColumn('failure_id', pd.Serial(not_null=True)),
        sql.Column('timestamp', pd.DateTime(not_null=True)),
        sql.Column('login', pd.String(not_null=True)),
        sql.Column('auth_type', pd.String(not_null=True)),
        sql.Column('ip_address', pd.String(not_null=True)),
        sql.Column('user_agent', pd.String()),
    )


class cms_pages(CommonAccesRights, Base_CachingTable):
    name = 'cms_pages'
    fields = (
        sql.PrimaryColumn('page_id', pd.Serial(not_null=True)),
        sql.Column('site', pd.String(not_null=True),
                   references=sql.a(sql.r.cms_config.site,
                                    onupdate='CASCADE', ondelete='CASCADE')),
        sql.Column('kind', pd.String(not_null=True)),
        sql.Column('identifier', pd.String(not_null=True)),
        sql.Column('parent', pd.Integer(), references=sql.r.cms_pages),
        sql.Column('modname', pd.String()),
        sql.Column('menu_visibility', pd.String(not_null=True)),
        sql.Column('foldable', pd.Boolean(not_null=False)),
        sql.Column('ord', pd.Integer(not_null=True)),
        sql.Column('tree_order', pd.String()),
        sql.Column('owner', pd.Integer(), references=sql.r.users),
        sql.Column('read_role_id', pd.PgName(not_null=True), default='anyone',
                   references=sql.a(sql.r.roles, onupdate='CASCADE')),
        sql.Column('write_role_id', pd.PgName(not_null=True),
                   default='cms-content-admin',
                   references=sql.a(sql.r.roles, onupdate='CASCADE', ondelete='SET DEFAULT')),
    )
    unique = (('identifier', 'site',),)

    @property
    def index_columns(self):
        parent = sqlalchemy.func.coalesce(sql.c.cms_pages.parent, ival(0))
        return (sql.a('ord', parent, 'site', 'kind', unique=True),)


class cms_pages_update_order(sql.SQLPlFunction, sql.SQLTrigger):
    name = 'cms_pages_update_order'
    table = cms_pages
    events = ('insert', 'update',)
    position = 'before'


class cms_page_texts(CommonAccesRights, Base_CachingTable):
    name = 'cms_page_texts'
    fields = (
        sql.Column('page_id', pd.Integer(not_null=True),
                   references=sql.a(sql.r.cms_pages, ondelete='CASCADE')),
        sql.Column('lang', pd.String(minlen=2, maxlen=2, not_null=True),
                   references=sql.a(sql.r.cms_languages.lang, onupdate='CASCADE')),
        sql.Column('published', pd.Boolean(not_null=True), default=True),
        sql.Column('parents_published', pd.Boolean(not_null=True)),
        sql.Column('creator', pd.Integer(not_null=True), references=sql.r.users),
        sql.Column('created', pd.DateTime(not_null=True), default=func.now()),
        sql.Column('published_since', pd.DateTime()),
        sql.Column('title', pd.String(not_null=True)),
        sql.Column('description', pd.String()),
        sql.Column('content', pd.String()),
        sql.Column('_title', pd.String()),
        sql.Column('_description', pd.String()),
        sql.Column('_content', pd.String()),
    )
    unique = (('page_id', 'lang',),)


class cms_page_tree_order(sql.SQLFunction):
    name = 'cms_page_tree_order'
    arguments = (sql.Column('page_id', pd.Integer()),)
    result_type = pd.String()
    depends_on = (cms_pages,)
    stability = 'stable'


class cms_page_tree_published(sql.SQLFunction):
    name = 'cms_page_tree_published'
    arguments = (sql.Column('page_id', pd.Integer()),
                 sql.Column('lang', pd.String()),)
    result_type = pd.Boolean()
    depends_on = (cms_pages, cms_page_texts)
    stability = 'stable'


class cms_v_pages(CommonAccesRights, SQLView):
    name = 'cms_v_pages'
    primary_column = 'page_key'

    @classmethod
    def query(cls):
        p = sql.t.cms_pages.alias('p')
        lang = sql.t.cms_languages.alias('l')
        t = sql.t.cms_page_texts.alias('t')
        cu = sql.t.users.alias('cu')
        ou = sql.t.users.alias('ou')
        return select([(stype(p.c.page_id) + sval('.') + lang.c.lang).label('page_key'),
                       p.c.site, p.c.kind, lang.c.lang,
                       p.c.page_id, p.c.identifier, p.c.parent, p.c.modname,
                       p.c.menu_visibility, p.c.foldable, p.c.ord, p.c.tree_order,
                       p.c.owner, p.c.read_role_id, p.c.write_role_id,
                       coalesce(t.c.published, False).label('published'),
                       coalesce(t.c.parents_published, False).label('parents_published'),
                       t.c.published_since,
                       t.c.creator, t.c.created,
                       coalesce(t.c.title, p.c.identifier).label('title_or_identifier'),
                       t.c.title, t.c.description, t.c.content, t.c._title,
                       t.c._description, t.c._content,
                       cu.c.login.label('creator_login'),
                       cu.c.user_.label('creator_name'),
                       ou.c.login.label('owner_login'),
                       ou.c.user_.label('owner_name'),
                       ],
                      from_obj=[p.
                                join(lang, ival(1) == 1).  # cross join
                                outerjoin(t, and_(t.c.page_id == p.c.page_id,
                                                  t.c.lang == lang.c.lang)).
                                outerjoin(cu, cu.c.uid == t.c.creator).
                                outerjoin(ou, ou.c.uid == p.c.owner)
                                ],
                      )

    def on_insert(self):
        pages = sql.t.cms_pages
        texts = sql.t.cms_page_texts
        new = self._reference('new')
        return (
            pages.insert().inline().values(**{c.name: getattr(new, c.name) for c in pages.c
                                              if c.name not in ('page_id', 'tree_order')}),
            pages.update().values(tree_order=func.cms_page_tree_order(pages.c.page_id)).where(
                and_(pages.c.site == new.site,
                     or_(pages.c.identifier == new.identifier,
                         pages.c.tree_order != func.cms_page_tree_order(pages.c.page_id)))
            ),
            texts.insert().from_select(
                texts.c,
                select(*[
                    pages.c.page_id if c.name == 'page_id' else getattr(new, c.name)
                    for c in texts.c
                ]).select_from(
                    pages
                ).where(and_(
                    pages.c.identifier == new.identifier,
                    pages.c.site == new.site,
                    pages.c.kind == new.kind,
                ))
            ).returning(*[
                (stype(texts.c.page_id) + sval('.') + texts.c.lang).label(c.name)
                if c.name == 'page_key' else
                getattr(texts.c, c.name) if hasattr(texts.c, c.name) else
                sqlalchemy.cast(null, c.type).label(c.name)
                for c in self.c
            ]),

        )

    def on_update(self):
        pages = sql.t.cms_pages
        texts = sql.t.cms_page_texts
        new = self._reference('new')
        old = self._reference('old')
        published = select(
            pages.c.page_id, texts.c.lang,
            func.cms_page_tree_published(pages.c.parent, texts.c.lang).label('parents_published')
        ).select_from(pages.join(texts, texts.c.page_id == pages.c.page_id)).alias('published')
        return (
            # Set the ord=0 first to work around the problem with recursion order in
            # cms_pages_update_order trigger (see the comment there for more info).
            pages.update().values(
                ord=ival(0)
            ).where(and_(
                pages.c.page_id == old.page_id,
                new.ord < old.ord,
            )),
            # Update the page structure and global data.
            pages.update().values(**{
                c.name: getattr(new, c.name) for c in pages.c
            }).where(
                pages.c.page_id == old.page_id
            ),
            # Update tree order for the whole site in case page structure has changed.
            pages.update().values(
                tree_order=func.cms_page_tree_order(pages.c.page_id)
            ).where(and_(
                pages.c.site == new.site,
                pages.c.tree_order != func.cms_page_tree_order(pages.c.page_id)
            )),
            # Update texts if they already exist for given page in given lang.
            texts.update().values(**{
                c.name: getattr(new, c.name)
                for c in texts.c if c.name not in ('page_id', 'lang', 'parents_published')
            }).where(and_(
                texts.c.page_id == old.page_id,
                texts.c.lang == new.lang
            )),
            # Insert texts if they didn't exist for given page in given lang.
            texts.insert().from_select(
                texts.c,
                select(*[
                    getattr(old if c.name == 'page_id' else new, c.name) for c in texts.c
                ]).where(and_(
                    new.lang.not_in(select(texts.c.lang).select_from(texts).where(texts.c.page_id==old.page_id)),
                    func.coalesce(new.title, new.description, new.content,
                                  new._title, new._description, new._content) != null,
                ))
            ),
            texts.update().values(
                parents_published=published.c.parents_published
            ).where(and_(
                texts.c.page_id == published.c.page_id,
                texts.c.lang == published.c.lang,
                texts.c.parents_published != published.c.parents_published,
            )),
        )

    delete_order = (cms_pages,)
    depends_on = (cms_pages, cms_page_texts, cms_page_tree_published,)


class cms_page_history(CommonAccesRights, sql.SQLTable):
    name = 'cms_page_history'
    fields = (
        sql.PrimaryColumn('history_id', pd.Serial(not_null=True)),
        sql.Column('page_id', pd.Integer(not_null=True)),
        sql.Column('lang', pd.String(minlen=2, maxlen=2, not_null=True)),
        sql.Column('uid', pd.Integer(not_null=True), references=sql.r.users),
        sql.Column('timestamp', pd.DateTime(not_null=True)),
        sql.Column('content', pd.String()),
        sql.Column('comment', pd.String()),
        sql.Column('inserted_lines', pd.Integer(not_null=True)),
        sql.Column('changed_lines', pd.Integer(not_null=True)),
        sql.Column('deleted_lines', pd.Integer(not_null=True)),
    )
    foreign_keys = (sql.a(('page_id', 'lang',),
                          (sql.r.cms_page_texts.page_id, sql.r.cms_page_texts.lang,),
                          ondelete='cascade'),)


class cms_v_page_history(CommonAccesRights, SQLView):
    name = 'cms_v_page_history'

    @classmethod
    def query(cls):
        h = sql.t.cms_page_history.alias('h')
        u = sql.t.users.alias('u')
        return select([h,
                       u.c.user_.label('user'),
                       u.c.login,
                       (stype(h.c.page_id) + sval('.') + h.c.lang).label('page_key'),
                       (stype(h.c.inserted_lines) + sval(' / ') +
                        stype(h.c.changed_lines) + sval(' / ') +
                        stype(h.c.deleted_lines)).label('changes')],
                      from_obj=[h.join(u, h.c.uid == u.c.uid)])
    insert_order = (cms_page_history,)


class cms_page_excerpts(sql.SQLTable):
    """Excerpts from CMS pages.
    Currently serving for printing parts of e-books in Braille.
    """
    name = 'cms_page_excerpts'
    fields = (
        sql.PrimaryColumn('id', pd.Serial(not_null=True)),
        sql.Column('page_id', pd.Integer(),
                   references=sql.a(sql.r.cms_pages, onupdate='CASCADE', ondelete='CASCADE')),
        sql.Column('lang', pd.String(not_null=True)),
        sql.Column('title', pd.String(not_null=True)),
        sql.Column('content', pd.String(not_null=True)),
    )
    access_rights = (('ALL', 'www-data',),)

#


class cms_page_attachments(CommonAccesRights, sql.SQLTable):
    name = 'cms_page_attachments'
    fields = (
        sql.PrimaryColumn('attachment_id', pd.Serial(not_null=True)),
        sql.Column('page_id', pd.Integer(not_null=True),
                   references=sql.a(sql.r.cms_pages, ondelete='CASCADE')),
        sql.Column('filename', pd.String(not_null=True)),
        sql.Column('mime_type', pd.String(not_null=True)),
        sql.Column('bytesize', pd.Integer(not_null=True)),
        sql.Column('created', pd.DateTime(not_null=True)),
        sql.Column('last_modified', pd.DateTime(not_null=True)),
        sql.Column('image', pd.Binary(),
                   doc="Resized image"),
        sql.Column('image_width', pd.Integer(),
                   doc="Resized image pixel width."),
        sql.Column('image_height', pd.Integer(),
                   doc="Resized image pixel height."),
        sql.Column('thumbnail', pd.Binary()),
        sql.Column('thumbnail_size', pd.String(),
                   doc="Desired thumbnail size - small/medium/large"),
        sql.Column('thumbnail_width', pd.Integer(),
                   doc="Thumbnail pixel width."),
        sql.Column('thumbnail_height', pd.Integer(),
                   doc="Thumbnail pixel height."),
        sql.Column('in_gallery', pd.Boolean(not_null=True), default=False),
        sql.Column('listed', pd.Boolean(not_null=True), default=False),
        sql.Column('author', pd.String()),
        sql.Column('location', pd.String()),
        sql.Column('width', pd.Integer()),
        sql.Column('height', pd.Integer()),
    )
    unique = (('filename', 'page_id',),)


class cms_page_attachment_texts(CommonAccesRights, sql.SQLTable):
    name = 'cms_page_attachment_texts'
    fields = (
        sql.Column('attachment_id', pd.Integer(not_null=True),
                   references=sql.a(sql.r.cms_page_attachments, ondelete='CASCADE',
                                    initially='DEFERRED')),
        sql.Column('lang', pd.String(minlen=2, maxlen=2, not_null=True),
                   references=sql.a(sql.r.cms_languages.lang,
                                    onupdate='CASCADE', ondelete='CASCADE')),
        sql.Column('title', pd.String()),
        sql.Column('description', pd.String()),
    )
    unique = (('attachment_id', 'lang',),)


class cms_v_page_attachments(CommonAccesRights, SQLView):
    name = 'cms_v_page_attachments'

    @classmethod
    def query(cls):
        a = sql.t.cms_page_attachments.alias('a')
        lang = sql.t.cms_languages.alias('l')
        t = sql.t.cms_page_attachment_texts.alias('t')
        return select([(stype(a.c.attachment_id) + sval('.') + lang.c.lang).label('attachment_key'),
                       lang.c.lang,
                       a.c.attachment_id, a.c.page_id, t.c.title, t.c.description,
                       a.c.filename, a.c.mime_type, a.c.bytesize,
                       a.c.image, a.c.image_width, a.c.image_height,
                       a.c.thumbnail, a.c.thumbnail_size, a.c.thumbnail_width, a.c.thumbnail_height,
                       a.c.in_gallery, a.c.listed, a.c.author, a.c.location, a.c.width, a.c.height,
                       a.c.created, a.c.last_modified],
                      from_obj=[a.join(lang, ival(1) == 1).  # cross join
                                outerjoin(t, and_(a.c.attachment_id == t.c.attachment_id,
                                                  lang.c.lang == t.c.lang))]
                      )

    def on_insert(self):
        return ("""(
    insert into cms_page_attachment_texts (attachment_id, lang, title, description)
           select new.attachment_id, new.lang, new.title, new.description
           where new.title is not null OR new.description is not null;
    insert into cms_page_attachments (attachment_id, page_id, filename, mime_type, bytesize,
                                 image, image_width, image_height,
                                 thumbnail, thumbnail_size, thumbnail_width, thumbnail_height,
                                 in_gallery, listed, author, "location", width, height,
                                 created, last_modified)
           values (new.attachment_id, new.page_id, new.filename, new.mime_type,
                   new.bytesize, new.image, new.image_width, new.image_height,
                   new.thumbnail, new.thumbnail_size,
                   new.thumbnail_width, new.thumbnail_height, new.in_gallery, new.listed,
                   new.author, new."location", new.width, new.height,
                   new.created, new.last_modified)
           returning
             attachment_id ||'.'|| (select max(lang) from cms_page_attachment_texts
                                    where attachment_id=attachment_id), null::char(2),
             attachment_id, page_id, null::text, null::text,
             filename, mime_type, bytesize, image, image_width, image_height, thumbnail,
             thumbnail_size, thumbnail_width, thumbnail_height, in_gallery, listed,
             author, "location", width, height, created, last_modified
        )""",)

    def on_update(self):
        return ("""(
    update cms_page_attachments set
           page_id = new.page_id,
           filename = new.filename,
           mime_type = new.mime_type,
           bytesize = new.bytesize,
           image = new.image,
           image_width = new.image_width,
           image_height = new.image_height,
           thumbnail = new.thumbnail,
           thumbnail_size = new.thumbnail_size,
           thumbnail_width = new.thumbnail_width,
           thumbnail_height = new.thumbnail_height,
           listed = new.listed,
           in_gallery = new.in_gallery,
           author = new.author,
           "location" = new."location",
           width = new.width,
           height = new.height,
           created = new.created,
           last_modified = new.last_modified
           where attachment_id = old.attachment_id;
    update cms_page_attachment_texts set
           title=new.title,
           description=new.description
           where attachment_id = old.attachment_id and lang = old.lang;
    insert into cms_page_attachment_texts (attachment_id, lang, title, description)
           select new.attachment_id, new.lang, new.title, new.description
           where old.attachment_id not in
             (select attachment_id from cms_page_attachment_texts where lang=old.lang);
        )""",)

    def on_delete(self):
        return ("""(
     delete from cms_page_attachments where attachment_id = old.attachment_id;
        )""",)


class cms_attachments_after_update_trigger(sql.SQLPlFunction, sql.SQLTrigger):
    name = 'cms_attachments_after_update_trigger'
    table = cms_page_attachments
    events = ('update',)
    position = 'after'

#


class cms_publications(CommonAccesRights, sql.SQLTable):
    """bibliographic data of the original (paper) books"""
    name = 'cms_publications'
    fields = (
        sql.PrimaryColumn('page_id', pd.Integer(not_null=True),
                          references=sql.a(sql.r.cms_pages, ondelete='CASCADE')),
        sql.Column('author', pd.String(not_null=True),
                   doc="creator(s) of the original work; full name(s), one name per line"),
        sql.Column('contributor', pd.String(),
                   doc=("creator(s) of the original work with less significant "
                        "role than author(s); full name(s), one name per line")),
        sql.Column('illustrator', pd.String(),
                   doc=("author(s) of illustrations in the original work; "
                        "full name(s), one name per line")),
        sql.Column('publisher', pd.String(),
                   doc="full name of the publisher"),
        sql.Column('published_year', pd.Integer(),
                   doc="year published"),
        sql.Column('edition', pd.Integer(),
                   doc="first, second, ..."),
        sql.Column('original_isbn', pd.String(),
                   doc=("ISBN identifier of the original work; if not null, the "
                        "publication is a work derived from it and fields author, "
                        "contributor, illustrator, publisher, published_year and "
                        "edition relate to the original work")),
        sql.Column('isbn', pd.String(),
                   'ISBN of this publication if it has one assigned'),
        sql.Column('uuid', pd.String(),
                   'Universally Unique Identifier if ISBN is not assigned'),
        sql.Column('adapted_by', pd.String(),
                   doc=("people or organization(s), who created this digital "
                        "publication if these are not already mentioned in the "
                        "above fields; full name(s), one name per line")),
        sql.Column('adapted_for', pd.String(),
                   doc=("organization or project for which this adaptation of the original "
                        "work into a digital publication has been done.")),
        sql.Column('cover_image', pd.Integer(),
                   references=sql.a(sql.r.cms_page_attachments, ondelete='SET NULL')),
        sql.Column('copyright_notice', pd.String()),
        sql.Column('notes', pd.String(),
                   doc="any other additional info, such as translator(s), reviewer(s) etc."),
        sql.Column('download_role_id', pd.PgName(),
                   references=sql.a(sql.r.roles, onupdate='CASCADE'),
                   doc="role allowed to download the offline version of the publication."),
    )


class cms_v_publications(CommonAccesRights, SQLView):
    name = 'cms_v_publications'
    primary_column = 'page_key'

    @classmethod
    def query(cls):
        pages = sql.t.cms_v_pages.alias('pages')
        publications = sql.t.cms_publications.alias('publications')
        attachments = sql.t.cms_page_attachments.alias('attachments')
        return select(
            cls._exclude(pages, pages.c.page_id) +
            cls._exclude(publications) +
            [attachments.c.filename.label('cover_image_filename')]
        ).select_from(
            pages
            .join(publications, publications.c.page_id == pages.c.page_id)
            .outerjoin(attachments, attachments.c.attachment_id == publications.c.cover_image)
        )

    def on_insert(self):
        pages = sql.t.cms_pages
        vpages = sql.t.cms_v_pages
        publications = sql.t.cms_publications
        texts = sql.t.cms_page_texts.alias('texts')
        new = self._reference('new')
        return (
            vpages.insert().values(**{
                c.name: getattr(new, c.name) for c in vpages.c if c.name not in ('page_id', 'page_key')
            }),
            publications.insert().from_select(
                publications.c,
                select([
                    c if hasattr(vpages.c, c.name) else getattr(new, c.name) for c in publications.c
                ]).select_from(
                    pages
                ).where(and_(
                    pages.c.identifier == new.identifier,
                    pages.c.site == new.site,
                    pages.c.kind == new.kind,
                )),
            ).returning(*[
                (stype(publications.c.page_id) + sval('.') +
                 select(func.min(texts.c.lang)).select_from(texts).where(
                     texts.c.page_id == publications.c.page_id
                 ).as_scalar()).label(c.name)
                if c.name == 'page_key' else
                getattr(publications.c, c.name) if hasattr(publications.c, c.name) else
                sqlalchemy.cast(null, c.type).label(c.name)
                for c in self.c
            ]),
        )

    update_order = (cms_v_pages, cms_publications)
    no_update_columns = ('page_id', 'page_key', 'lang')
    delete_order = (cms_v_pages,)


class cms_publication_languages(CommonAccesRights, sql.SQLTable):
    """list of content languages available for given publication"""
    name = 'cms_publication_languages'
    fields = (
        sql.Column('page_id', pd.Integer(not_null=True),
                   references=sql.a(sql.r.cms_publications.page_id, ondelete='CASCADE')),
        sql.Column('lang', pd.String(not_null=True),
                   doc="language code"),
    )
    unique = (('page_id', 'lang',),)


class cms_publication_indexes(CommonAccesRights, sql.SQLTable):
    """list of indexes available for given publication"""
    name = 'cms_publication_indexes'
    fields = (
        sql.PrimaryColumn('index_id', pd.Serial(not_null=True)),
        sql.Column('page_id', pd.Integer(not_null=True),
                   references=sql.a(sql.r.cms_publications.page_id, ondelete='CASCADE')),
        sql.Column('title', pd.String(not_null=True)),
    )
    unique = (('page_id', 'title',),)


class cms_publication_exports(CommonAccesRights, sql.SQLTable):
    """Exported publication versions."""
    name = 'cms_publication_exports'
    fields = (
        sql.PrimaryColumn('export_id', pd.Serial()),
        sql.Column('page_id', pd.Integer(not_null=True),
                   references=sql.a(sql.r.cms_publications.page_id, ondelete='CASCADE')),
        sql.Column('lang', pd.String(minlen=2, maxlen=2, not_null=True),
                   references=sql.a(sql.r.cms_languages.lang,
                                    onupdate='CASCADE', ondelete='CASCADE')),
        sql.Column('format', pd.String(not_null=True)),
        sql.Column('version', pd.String(not_null=True)),
        sql.Column('timestamp', pd.DateTime(not_null=True)),
        sql.Column('public', pd.Boolean(not_null=True), default=True),
        sql.Column('bytesize', pd.Integer(not_null=True)),
        sql.Column('notes', pd.String()),
        sql.Column('log', pd.String()),
    )
    unique = (('page_id', 'lang', 'format', 'version',),)


class cms_v_publication_exports(CommonAccesRights, SQLView):
    name = 'cms_v_publication_exports'

    @classmethod
    def query(cls):
        exports = sql.t.cms_publication_exports.alias('e')
        return select(list(exports.c) +
                      [(stype(exports.c.page_id) + sval('.') + exports.c.lang).label('page_key')],
                      from_obj=exports)
    insert_order = (cms_publication_exports,)
    update_order = (cms_publication_exports,)
    delete_order = (cms_publication_exports,)


class cms_news(CommonAccesRights, Base_CachingTable):
    name = 'cms_news'
    fields = (
        sql.PrimaryColumn('news_id', pd.Serial(not_null=True)),
        sql.Column('page_id', pd.Integer(not_null=True),
                   references=sql.a(sql.r.cms_pages, ondelete='CASCADE')),
        sql.Column('lang', pd.String(minlen=2, maxlen=2, not_null=True),
                   references=sql.a(sql.r.cms_languages.lang, onupdate='CASCADE')),
        sql.Column('author', pd.Integer(not_null=True),
                   references=sql.r.users),
        sql.Column('timestamp', pd.DateTime(not_null=True), default=func.now()),
        sql.Column('title', pd.String(not_null=True)),
        sql.Column('content', pd.String(not_null=True)),
        sql.Column('days_displayed', pd.Integer(not_null=True)),
    )


class cms_v_news(CommonAccesRights, SQLView):
    name = 'cms_v_news'

    @classmethod
    def query(cls):
        n = sql.t.cms_news.alias('n')
        u = sql.t.users.alias('u')
        return select([n,
                       u.c.user_.label('author_name'),
                       u.c.login.label('author_login'),
                       ],
                      from_obj=[n.join(u, n.c.author == u.c.uid)])
    insert_order = (cms_news,)
    update_order = (cms_news,)
    delete_order = (cms_news,)
    no_insert_columns = ('news_id',)


class cms_recent_timestamp(sql.SQLFunction):
    """Return true if `ts' is not older than `max_days' days.  Needed for a pytis
    filtering condition (FunctionCondition is currently too simple to express
    this directly).
    """
    name = 'cms_recent_timestamp'
    arguments = (sql.Column('ts', pd.DateTime()),
                 sql.Column('max_days', pd.Integer()),)
    result_type = pd.Boolean()
    stability = 'stable'

    def body(self):
        """Return true if `ts' is not older than `max_days' days.  Needed for a pytis
        filtering condition (FunctionCondition is currently too simple to express
        this directly).
        """
        return 'select (current_date - $1::date) < $2'


class cms_planner(CommonAccesRights, Base_CachingTable):
    name = 'cms_planner'
    fields = (
        sql.PrimaryColumn('planner_id', pd.Serial(not_null=True)),
        sql.Column('page_id', pd.Integer(not_null=True),
                   references=sql.a(sql.r.cms_pages, ondelete='CASCADE')),
        sql.Column('lang', pd.String(minlen=2, maxlen=2, not_null=True),
                   references=sql.a(sql.r.cms_languages.lang, onupdate='CASCADE')),
        sql.Column('author', pd.Integer(not_null=True),
                   references=sql.r.users),
        sql.Column('timestamp', pd.DateTime(not_null=True), default=func.now()),
        sql.Column('start_date', pd.Date(not_null=True)),
        sql.Column('end_date', pd.Date()),
        sql.Column('title', pd.String(not_null=True)),
        sql.Column('content', pd.String(not_null=True)),
    )


class cms_v_planner(CommonAccesRights, SQLView):
    name = 'cms_v_planner'

    @classmethod
    def query(cls):
        p = sql.t.cms_planner.alias('p')
        u = sql.t.users.alias('u')
        return select([p,
                       u.c.user_.label('author_name'),
                       u.c.login.label('author_login'),
                       ],
                      from_obj=[p.join(u, p.c.author == u.c.uid)])
    insert_order = (cms_planner,)
    update_order = (cms_planner,)
    delete_order = (cms_planner,)
    no_insert_columns = ('planner_id',)


class cms_newsletters(CommonAccesRights, Base_CachingTable):
    name = 'cms_newsletters'
    fields = (
        sql.PrimaryColumn('newsletter_id', pd.Serial(not_null=True)),
        sql.Column('page_id', pd.Integer(not_null=True),
                   references=sql.a(sql.r.cms_pages, ondelete='CASCADE')),
        sql.Column('lang', pd.String(minlen=2, maxlen=2, not_null=True),
                   references=sql.a(sql.r.cms_languages.lang, onupdate='CASCADE')),
        sql.Column('title', pd.String(not_null=True)),
        sql.Column('image', pd.Binary(not_null=True)),
        sql.Column('image_width', pd.Integer(not_null=True)),
        sql.Column('image_height', pd.Integer(not_null=True)),
        sql.Column('description', pd.String(not_null=True)),
        sql.Column('sender', pd.String(not_null=True)),
        sql.Column('address', pd.String(not_null=True)),
        sql.Column('read_role_id', pd.PgName(not_null=True), default='anyone',
                   references=sql.a(sql.r.roles, onupdate='CASCADE')),
        sql.Column('write_role_id', pd.PgName(not_null=True), default='cms-content-admin',
                   references=sql.a(sql.r.roles, onupdate='CASCADE', ondelete='SET DEFAULT')),
        sql.Column('bg_color', pd.String(not_null=True)),
        sql.Column('text_color', pd.String(not_null=True)),
        sql.Column('link_color', pd.String(not_null=True)),
        sql.Column('heading_color', pd.String(not_null=True)),
        sql.Column('top_bg_color', pd.String(not_null=True)),
        sql.Column('top_text_color', pd.String(not_null=True)),
        sql.Column('top_link_color', pd.String(not_null=True)),
        sql.Column('footer_bg_color', pd.String(not_null=True)),
        sql.Column('footer_text_color', pd.String(not_null=True)),
        sql.Column('footer_link_color', pd.String(not_null=True)),
        sql.Column('signature', pd.String()),
    )


class cms_newsletter_subscription(CommonAccesRights, Base_CachingTable):
    name = 'cms_newsletter_subscription'
    fields = (
        sql.PrimaryColumn('subscription_id', pd.Serial(not_null=True)),
        sql.Column('newsletter_id', pd.Integer(not_null=True),
                   references=sql.a(sql.r.cms_newsletters, ondelete='CASCADE')),
        sql.Column('uid', pd.Integer(), references=sql.r.users),
        sql.Column('email', pd.String()),
        sql.Column('code', pd.String(), doc='Unsubscription code (email only)'),
        sql.Column('timestamp', pd.DateTime(not_null=True), default=func.now()),
    )
    check = ('(uid is not null or email is not null and code is not null) '
             'and (uid is null or email is null and code is null)',)
    unique = (('newsletter_id', 'uid',),
              ('newsletter_id', 'email',),)


class cms_v_newsletter_subscription(CommonAccesRights, SQLView):
    name = 'cms_v_newsletter_subscription'

    @classmethod
    def query(cls):
        s = sql.t.cms_newsletter_subscription.alias('s')
        u = sql.t.users.alias('u')
        return select(cls._exclude(s, s.c.email) +
                      [coalesce(s.c.email, u.c.email).label('email'),
                       u.c.user_.label('user_name'),
                       u.c.login.label('user_login'),
                       ],
                      from_obj=[s.outerjoin(u, s.c.uid == u.c.uid)])

    def on_insert(self):
        return ("insert into public.cms_newsletter_subscription "
                "(newsletter_id, email, uid, code, timestamp) "
                "VALUES (new.newsletter_id, new.email, new.uid, new.code, new.timestamp);",)
    delete_order = (cms_newsletter_subscription,)
    no_insert_columns = ('subscription_id',)


class cms_newsletter_editions(CommonAccesRights, Base_CachingTable):
    name = 'cms_newsletter_editions'
    fields = (
        sql.PrimaryColumn('edition_id', pd.Serial(not_null=True)),
        sql.Column('newsletter_id', pd.Integer(not_null=True),
                   references=sql.a(sql.r.cms_newsletters, ondelete='CASCADE')),
        sql.Column('creator', pd.Integer(not_null=True), references=sql.r.users),
        sql.Column('created', pd.DateTime(not_null=True), default=func.now()),
        sql.Column('sent', pd.DateTime()),
        sql.Column('access_code', pd.String()),
    )


class cms_newsletter_posts(CommonAccesRights, Base_CachingTable):
    name = 'cms_newsletter_posts'
    fields = (
        sql.PrimaryColumn('post_id', pd.Serial(not_null=True)),
        sql.Column('edition_id', pd.Integer(not_null=True),
                   references=sql.a(sql.r.cms_newsletter_editions, ondelete='CASCADE')),
        sql.Column('ord', pd.Integer()),
        sql.Column('title', pd.String(not_null=True)),
        sql.Column('content', pd.String(not_null=True)),
        sql.Column('image', pd.Binary()),
        sql.Column('image_position', pd.String()),
        sql.Column('image_width', pd.Integer()),
        sql.Column('image_height', pd.Integer()),
    )


class cms_discussions(CommonAccesRights, sql.SQLTable):
    name = 'cms_discussions'
    fields = (
        sql.PrimaryColumn('comment_id', pd.Serial(not_null=True)),
        sql.Column('page_id', pd.Integer(not_null=True),
                   references=sql.a(sql.r.cms_pages, ondelete='CASCADE')),
        sql.Column('lang', pd.String(minlen=2, maxlen=2, not_null=True),
                   references=sql.a(sql.r.cms_languages.lang, onupdate='CASCADE')),
        sql.Column('author', pd.Integer(not_null=True),
                   references=sql.r.users),
        sql.Column('timestamp', pd.DateTime(not_null=True), default=func.now()),
        sql.Column('in_reply_to', pd.Integer(),
                   references=sql.a(sql.r.cms_discussions, ondelete='SET NULL')),
        sql.Column('tree_order', pd.String(not_null=True), index=True),
        sql.Column('text', pd.String(not_null=True)),
    )


class cms_discussions_trigger_before_insert(sql.SQLPlFunction, sql.SQLTrigger):
    name = 'cms_discussions_trigger_before_insert'
    arguments = ()
    table = cms_discussions
    position = 'before'
    events = ('insert',)


class cms_panels(CommonAccesRights, Base_CachingTable):
    name = 'cms_panels'
    fields = (
        sql.PrimaryColumn('panel_id', pd.Serial(not_null=True)),
        sql.Column('site', pd.String(not_null=True),
                   references=sql.a(sql.r.cms_config.site,
                                    onupdate='CASCADE', ondelete='CASCADE')),
        sql.Column('lang', pd.String(minlen=2, maxlen=2, not_null=True),
                   references=sql.a(sql.r.cms_languages.lang, onupdate='CASCADE')),
        sql.Column('identifier', pd.String()),
        sql.Column('title', pd.String(not_null=True)),
        sql.Column('ord', pd.Integer()),
        sql.Column('page_id', pd.Integer(),
                   references=sql.a(sql.r.cms_pages, ondelete='SET NULL')),
        sql.Column('size', pd.Integer()),
        sql.Column('content', pd.String()),
        sql.Column('_content', pd.String()),
        sql.Column('published', pd.Boolean(not_null=True), default=False),
    )
    unique = (('identifier', 'site', 'lang',),)


class cms_v_panels(CommonAccesRights, SQLView):
    name = 'cms_v_panels'

    @classmethod
    def query(cls):
        cms_panels = sql.t.cms_panels
        cms_pages = sql.t.cms_pages
        return select([cms_panels, cms_pages.c.modname, cms_pages.c.read_role_id],
                      from_obj=[cms_panels.
                                outerjoin(cms_pages, cms_panels.c.page_id == cms_pages.c.page_id)])
    insert_order = (cms_panels,)
    update_order = (cms_panels,)
    delete_order = (cms_panels,)

#


class cms_stylesheets(CommonAccesRights, Base_CachingTable):
    name = 'cms_stylesheets'
    fields = (
        sql.PrimaryColumn('stylesheet_id', pd.Serial(not_null=True)),
        sql.Column('site', pd.String(not_null=True),
                   references=sql.a(sql.r.cms_config.site,
                                    onupdate='CASCADE', ondelete='CASCADE')),
        sql.Column('filename', pd.String(maxlen=32, not_null=True)),
        sql.Column('active', pd.Boolean(not_null=True), default=True),
        sql.Column('scope', pd.String()),
        sql.Column('description', pd.String()),
        sql.Column('content', pd.String()),
        sql.Column('ord', pd.Integer()),
    )
    unique = (('filename', 'site',),)
    init_columns = ('filename', 'site', 'ord',)
    init_values = (
        ('default.css', '*', 10),
    )


class cms_themes(CommonAccesRights, Base_CachingTable):
    name = 'cms_themes'
    fields = (
        sql.PrimaryColumn('theme_id', pd.Serial(not_null=True),
                          references=sql.r.cms_themes),
        sql.Column('name', pd.String(not_null=True), unique=True),
        sql.Column('foreground', pd.String(maxlen=7)),
        sql.Column('background', pd.String(maxlen=7)),
        sql.Column('border', pd.String(maxlen=7)),
        sql.Column('heading_fg', pd.String(maxlen=7)),
        sql.Column('heading_bg', pd.String(maxlen=7)),
        sql.Column('heading_line', pd.String(maxlen=7)),
        sql.Column('frame_fg', pd.String(maxlen=7)),
        sql.Column('frame_bg', pd.String(maxlen=7)),
        sql.Column('frame_border', pd.String(maxlen=7)),
        sql.Column('link', pd.String(maxlen=7)),
        sql.Column('link_visited', pd.String(maxlen=7)),
        sql.Column('link_hover', pd.String(maxlen=7)),
        sql.Column('meta_fg', pd.String(maxlen=7)),
        sql.Column('meta_bg', pd.String(maxlen=7)),
        sql.Column('help', pd.String(maxlen=7)),
        sql.Column('error_fg', pd.String(maxlen=7)),
        sql.Column('error_bg', pd.String(maxlen=7)),
        sql.Column('error_border', pd.String(maxlen=7)),
        sql.Column('message_fg', pd.String(maxlen=7)),
        sql.Column('message_bg', pd.String(maxlen=7)),
        sql.Column('message_border', pd.String(maxlen=7)),
        sql.Column('table_cell', pd.String(maxlen=7)),
        sql.Column('table_cell2', pd.String(maxlen=7)),
        sql.Column('top_fg', pd.String(maxlen=7)),
        sql.Column('top_bg', pd.String(maxlen=7)),
        sql.Column('top_border', pd.String(maxlen=7)),
        sql.Column('highlight_bg', pd.String(maxlen=7)),
        sql.Column('inactive_folder', pd.String(maxlen=7)),
    )
    init_columns = (
        'name', 'foreground', 'background', 'border', 'heading_fg', 'heading_bg',
        'heading_line', 'frame_fg', 'frame_bg', 'frame_border', 'link',
        'link_visited', 'link_hover', 'meta_fg', 'meta_bg', 'help', 'error_fg',
        'error_bg', 'error_border', 'message_fg', 'message_bg', 'message_border',
        'table_cell', 'table_cell2', 'top_fg', 'top_bg', 'top_border',
        'highlight_bg', 'inactive_folder',)
    init_values = (
        ('Yellowstone', '#000', '#fff9ec', '#eda', '#420', '#fff0b0', '#eca',
         '#000', '#fff0d4', '#ffde90', '#a30', '#a30', '#f40', None, None,
         '#553', None, None, None, None, None, None, '#fff', '#fff8f0', '#444',
         '#fff', '#db9', '#fb7', '#ed9'),
        ('Olive', '#000', '#fff', '#bcb', '#0b4a44', '#d2e0d8', None, '#000',
         '#e8eee8', '#d0d7d0', '#042', None, '#d72', None, None, None, None,
         '#fc9', '#fa8', None, '#dfd', '#aea', '#f8fbfa', '#f1f3f2', None,
         '#efebe7', '#8a9', '#fc8', '#d2e0d8'),
    )


class cms_system_text_labels(CommonAccesRights, Base_CachingTable):
    name = 'cms_system_text_labels'
    fields = (
        sql.Column('label', pd.PgName(not_null=True)),
        sql.Column('site', pd.String(not_null=True),
                   references=sql.a(sql.r.cms_config.site,
                                    onupdate='CASCADE', ondelete='CASCADE')),
    )
    unique = (('label', 'site',),)


class cms_add_text_label(sql.SQLPlFunction):
    """"""
    name = 'cms_add_text_label'
    arguments = (
        sql.Column('_label', pd.PgName()),
        sql.Column('_site', pd.String()),
    )
    result_type = None


class cms_system_texts(CommonAccesRights, Base_CachingTable):
    name = 'cms_system_texts'
    fields = (
        sql.Column('label', pd.PgName(not_null=True)),
        sql.Column('site', pd.String(not_null=True)),
        sql.Column('lang', pd.String(minlen=2, maxlen=2, not_null=True),
                   references=sql.a(sql.r.cms_languages.lang,
                                    onupdate='CASCADE', ondelete='CASCADE')),
        sql.Column('description', pd.String(), default=''),
        sql.Column('content', pd.String(), default=''),
    )
    unique = (('label', 'site', 'lang',),)
    foreign_keys = (sql.a(('label', 'site',),
                          (sql.r.cms_system_text_labels.label, sql.r.cms_system_text_labels.site,),
                          onupdate='CASCADE', ondelete='CASCADE'),)


class cms_v_system_text_labels(CommonAccesRights, SQLView):
    name = 'cms_v_system_text_labels'

    @classmethod
    def query(cls):
        labels = sql.t.cms_system_text_labels
        languages = sql.t.cms_languages
        return select([labels.c.label, labels.c.site, languages.c.lang],
                      from_obj=[labels, languages])


class cms_v_system_texts(CommonAccesRights, SQLView):
    name = 'cms_v_system_texts'

    @classmethod
    def query(cls):
        labels = sql.t.cms_v_system_text_labels
        texts = sql.t.cms_system_texts
        return select(
            [(labels.c.label + sval(':') + labels.c.site + sval(':') + labels.c.lang)
             .label('text_id'),
             labels.c.label, labels.c.site, labels.c.lang, texts.c.description, texts.c.content],
            from_obj=[labels.outerjoin(texts, and_(labels.c.lang == texts.c.lang,
                                                   labels.c.label == texts.c.label,
                                                   labels.c.site == texts.c.site))],
        )

    def on_update(self):
        return ("""(
    delete from cms_system_texts where label = new.label and lang = new.lang and site = new.site;
    insert into cms_system_texts (label, site, lang, description, content)
           values (new.label, new.site, new.lang, new.description, new.content);
        )""",)
    # gensqlalchemy doesn't generate a proper where clause here (maybe multicolumn foreign key?)
    # delete_order = (cms_system_text_labels,)

    def on_delete(self):
        return ("delete from cms_system_text_labels where label = old.label and site = old.site;",)

#


class cms_email_labels(CommonAccesRights, sql.SQLTable):
    name = 'cms_email_labels'
    fields = (
        sql.PrimaryColumn('label', pd.PgName(not_null=True)),
    )


class cms_add_email_label(sql.SQLPlFunction):
    name = 'cms_add_email_label'
    arguments = (
        sql.Column('_label', pd.PgName()),
    )
    result_type = None


class cms_emails(CommonAccesRights, sql.SQLTable):
    name = 'cms_emails'
    fields = (
        sql.Column('label', pd.PgName(not_null=True),
                   references=sql.r.cms_email_labels),
        sql.Column('lang', pd.String(minlen=2, maxlen=2, not_null=True),
                   references=sql.a(sql.r.cms_languages.lang,
                                    onupdate='CASCADE', ondelete='CASCADE')),
        sql.Column('description', pd.String()),
        sql.Column('subject', pd.String()),
        sql.Column('cc', pd.String()),
        sql.Column('content', pd.String(), default=''),
    )
    unique = (('label', 'lang',),)


class cms_v_emails(CommonAccesRights, SQLView):
    name = 'cms_v_emails'

    @classmethod
    def query(cls):
        el = sql.c.cms_email_labels
        lang = sql.c.cms_languages
        e = sql.c.cms_emails
        return select([(el.label + sval(':') + lang.lang).label('text_id'),
                       el.label, lang.lang, e.description, e.subject, e.cc, e.content],
                      from_obj=[sql.t.cms_email_labels,
                                sql.t.cms_languages.outerjoin(sql.t.cms_emails)],
                      whereclause=and_(el.label == e.label, lang.lang == e.lang))

    def on_insert(self):
        return ("""(
    select cms_add_email_label(new.label);
    insert into cms_emails values (new.label, new.lang, new.description, new.subject,
                                   new.cc, new.content);
        )""",)

    def on_update(self):
        return ("""(
    delete from cms_emails where label = new.label and lang = new.lang;
    insert into cms_emails values (new.label, new.lang, new.description, new.subject,
                                   new.cc, new.content);
        )""",)

    def on_delete(self):
        return ("""(
    delete from cms_emails where label = old.label;
    delete from cms_email_labels where label = old.label;
        )""",)


class cms_email_attachments(CommonAccesRights, sql.SQLTable):
    name = 'cms_email_attachments'
    fields = (
        sql.PrimaryColumn('attachment_id', pd.Serial(not_null=True)),
        sql.Column('label', pd.PgName(not_null=True),
                   references=sql.a(sql.r.cms_email_labels, ondelete='CASCADE')),
        sql.Column('filename', pd.String(not_null=True)),
        sql.Column('mime_type', pd.String(not_null=True)),
    )


class cms_email_spool(CommonAccesRights, sql.SQLTable):
    name = 'cms_email_spool'
    fields = (
        sql.PrimaryColumn('id', pd.Serial(not_null=True)),
        sql.Column('sender_address', pd.String()),
        sql.Column('role_id', pd.PgName(),
                   references=sql.a(sql.r.roles, onupdate='CASCADE', ondelete='CASCADE'),
                   doc="recipient role, if NULL then all users"),
        sql.Column('subject', pd.String()),
        sql.Column('content', pd.String(),
                   doc="body of the e-mail"),
        sql.Column('date', pd.DateTime(), default=func.now(),
                   doc="time of insertion"),
        sql.Column('pid', pd.Integer()),
        sql.Column('finished', pd.Boolean(not_null=False), default=False,
                doc="set TRUE after the mail was successfully sent"),
    )


class cms_crypto_names(CommonAccesRights, sql.SQLTable):
    name = 'cms_crypto_names'
    fields = (
        sql.PrimaryColumn('name', pd.String(not_null=True)),
        sql.Column('description', pd.String()),
    )
    init_columns = ('name', 'description',)
    access_rights = (('ALL', 'www-data',),)


class cms_crypto_keys(CommonAccesRights, sql.SQLTable):
    name = 'cms_crypto_keys'
    fields = (
        sql.PrimaryColumn('key_id', pd.Serial(not_null=True)),
        sql.Column('name', pd.String(not_null=True),
                   references=sql.a(sql.r.cms_crypto_names,
                                    onupdate='CASCADE', ondelete='CASCADE')),
        sql.Column('uid', pd.Integer(not_null=True),
                   references=sql.a(sql.r.users, onupdate='CASCADE', ondelete='CASCADE')),
        sql.Column('key', pd.Binary(not_null=True)),
    )
    unique = (('name', 'uid',),)
    access_rights = (('ALL', 'www-data',),)


class cms_crypto_extract_key(sql.SQLPlFunction):
    name = 'cms_crypto_extract_key'
    arguments = (sql.Column('encrypted', pd.Binary()),
                 sql.Column('psw', pd.String()),)
    result_type = pd.String()
    stability = 'immutable'


class cms_crypto_store_key(sql.SQLPlFunction):
    "This is a PL/pgSQL, and not SQL, function in order to prevent direct dependency on pg_crypto."
    name = 'cms_crypto_store_key'
    arguments = (sql.Column('key', pd.String()),
                 sql.Column('psw', pd.String()),)
    result_type = pd.Binary()
    stability = 'immutable'

    def body(self):
        return "begin return pgp_sym_encrypt('wiking:'||$1, $2); end;"


class cms_crypto_insert_key(sql.SQLPlFunction):
    name = 'cms_crypto_insert_key'
    arguments = (sql.Column('name_', pd.String()),
                 sql.Column('uid_', pd.Integer()),
                 sql.Column('key_', pd.String()),
                 sql.Column('psw', pd.String()),)
    result_type = pd.Boolean()


class cms_crypto_change_password(sql.SQLPlFunction):
    name = 'cms_crypto_change_password'
    arguments = (sql.Column('id_', pd.Integer()),
                 sql.Column('old_psw', pd.String()),
                 sql.Column('new_psw', pd.String()),)
    result_type = pd.Boolean()


class cms_crypto_copy_key(sql.SQLPlFunction):
    name = 'cms_crypto_copy_key'
    arguments = (sql.Column('name_', pd.String()),
                 sql.Column('from_uid', pd.Integer()),
                 sql.Column('to_uid', pd.Integer()),
                 sql.Column('from_psw', pd.String()),
                 sql.Column('to_psw', pd.String()),)
    result_type = pd.Boolean()


class cms_crypto_delete_key(sql.SQLPlFunction):
    name = 'cms_crypto_delete_key'
    arguments = (sql.Column('name_', pd.String()),
                 sql.Column('uid_', pd.Integer()),
                 sql.Column('force', pd.Boolean()),)
    result_type = pd.Boolean()

#


class cms_crypto_unlocked_passwords(CommonAccesRights, sql.SQLTable):
    name = 'cms_crypto_unlocked_passwords'
    fields = (
        sql.Column('key_id', pd.Integer(not_null=True),
                   references=sql.a(sql.r.cms_crypto_keys,
                                    onupdate='CASCADE', ondelete='CASCADE')),
        sql.Column('password', pd.Binary()),
    )
    access_rights = (('ALL', 'www-data',),)


class cms_crypto_unlocked_passwords_insert_trigger(sql.SQLPlFunction, sql.SQLTrigger):
    name = 'cms_crypto_unlocked_passwords_insert_trigger'
    arguments = ()
    table = cms_crypto_unlocked_passwords
    position = 'before'
    events = ('insert',)


class cms_crypto_unlock_passwords(sql.SQLFunction):
    name = 'cms_crypto_unlock_passwords'
    arguments = (sql.Column('uid_', pd.Integer()),
                 sql.Column('psw', pd.String()),
                 sql.Column('cookie', pd.String()),)
    result_type = None
    depends_on = (cms_crypto_unlocked_passwords,)


class cms_crypto_lock_passwords(sql.SQLFunction):
    name = 'cms_crypto_lock_passwords'
    arguments = (sql.Column('uid_', pd.Integer()),)
    result_type = None
    depends_on = (cms_crypto_unlocked_passwords,)


class cms_crypto_cook_passwords(sql.SQLPlFunction):
    name = 'cms_crypto_cook_passwords'
    arguments = (sql.Column('uid_', pd.Integer()),
                 sql.Column('cookie', pd.String()),)
    result_type = pd.String()
    multirow = True


class pytis_crypto_unlock_current_user_passwords(sql.SQLFunction):
    """Dummy function to avoid error messages in logs (called by Pytis)."""
    name = 'pytis_crypto_unlock_current_user_passwords'
    arguments = (sql.Column('password_', pd.String()),)
    result_type = pd.String()
    multirow = True
    stability = 'immutable'

    def body(self):
        return "select ''::text where false;"


class pytis_crypto_db_key(sql.SQLFunction):
    """Dummy function needed by Pytis initialization when dbpass is set."""
    name = 'pytis_crypto_db_key'
    arguments = (
        sql.Column('key_name_', pd.String()),
    )
    result_type = pd.String()
    multirow = False
    stability = 'STABLE'
    security_definer = True

    def body(self):
        return "select null::text;"



# -*- coding: utf-8 -*-
#
# Copyright (C) 2009-2016 OUI Technology Ltd.
# Copyright (C) 2019 Tomáš Cerha <t.cerha@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import lcg
from wiking.cms import Text
from pytis.presentation import TextFormat

_ = lcg.TranslatableTextFactory('wiking-cms')

top = Text(
    'cms.top',
    _("Text displayed at the very top of every page next to the site title"),
    '')

disabled = Text(
    'cms.disabled',
    # Translators: Web form label (account in the meaning of user account)
    _("Information about a locked account"),
    # Translators: Text presented to a user in a web page
    _("Your account is disabled. Protected services are not available. "
      "Contact the administrators to get the account enabled."),
    text_format=TextFormat.PLAIN)

unconfirmed = Text(
    'cms.unconfirmed',
    # Translators: Web form label
    _("Information about an account waiting for activation code confirmation"),
    # Translators: Text presented to a user in a web page
    _("Your account has not been activated yet.  Please, check your e-mail and follow "
      "the provided link or enter the activation code into the [%(uri)s activation "
      "form].  Protected services will not be available until you perform "
      "activation."),
    text_format=TextFormat.PLAIN)

unapproved = Text(
    'cms.unapproved',
    # Translators: Web form label
    _("Information about an account waiting for admin approval"),
    # Translators: Text presented to a user in a web page
    _("Your account is waiting for approval by the administrator. "
      "The protected services are not available until the account gets "
      "approved. Contact the administrators if you feel it takes too long."),
    text_format=TextFormat.PLAIN)

regintro = Text(
    'cms.regintro',
    _("Introductory text displayed above the registration form"),
    None)

regsuccess = Text(
    'cms.regsuccess',
    _("Text displayed after successful user registration"),
    _("Your account now awaits administrator's approval."))

regsuccess_autoapproved = Text(
    'cms.regsuccess_autoapproved',
    _("Text displayed after successful user registration when autoapprove_new_users is set "
      "in configuration"),
    _("Your account is now fully functional. You can [%s log in] now.\n\n"
      "You may still need some further privileges to access certain restricted services.",
      '?command=login'))

regconfirm = Text(
    'cms.regconfirm',
    _("Information displayed above the checkbox for confirmation of site specific "
      "conditions during registration and in profile edit form (if empty, no confirmation "
      "is required)"),
    None)

personal_data_management = Text(
    'cms.personal_data_management',
    _("Information about personal data management (displayed in the registration "
      "form and user profile)."),
    _("Personal data which you are submitting in this form and which arise from your "
      "usage of this application are processed by its operator >>insert name and contact<< "
      "who also shares them for specifically delimited purposes with >>insert<<. "
      "You have a possibility to adjust your personal data at any time as well as "
      "to cancel your account and have your data deleted.  See >>insert link<< for "
      "more details."))

registration_note_descr = Text(
    'cms.registration_note_descr',
    _("Description of the field \"Note\" in the registration form (server reload required "
      "on change)."),
    _("Optional message for the administrator. If you wish to communicate something "
      "to the administrator processing your registration, you can do so here. This "
      "is not normally necessary."),
    text_format=TextFormat.PLAIN)

footer = Text(
    'cms.footer',
    _("Text displayed at the very bottom of every page"),
    # Translators: Label followed by an email address link.
    _("Contact:") + ' $webmaster_address')

default_copyright_notice = Text(
    'cms.default_copyright_notice',
    _("Default value of the Copyright Notice field for newly created publications."),
    None)

login_dialog_top_text = Text(
    'cms.login_dialog_top_text',
    _("Text displayed above the login dialog"),
    None)

login_dialog_bottom_text = Text(
    'cms.login_dialog_bottom_text',
    _("Text displayed below the login dialog"),
    None)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2006-2016 OUI Technology Ltd.
# Copyright (C) 2019-2025 Tomáš Cerha <t.cerha@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

__version__ = '2.2.1'

from .util import (  # noqa: F401
    DBG, EVT, OPR,
    Abort, AuthenticationError, AuthenticationProvider, AuthenticationRedirect,
    AuthorizationError, BadRequest, Binding, Channel, ChannelContent, ConfirmationDialog,
    CookieAuthenticationProvider, Date, DateTime, DecryptionDialog, Document,
    Forbidden, HTTPBasicAuthenticationProvider, HtmlContent, IFrame, InputForm,
    InternalServerError, LanguageSelection, LoginControl, LoginDialog, MailAttachment,
    MaximizedModeControl, MenuItem, Message, ModuleInstanceResolver, NotAcceptable, NotFound,
    NotModified, Panel, PasswordStorage,
    Pbkdf2Md5PasswordStorage, Pbkdf2PasswordStorage, PermanentRedirect, PlainTextPasswordStorage,
    Redirect, RequestError, Response, RssWriter, ServiceUnavailable,
    Specification, TZInfo, Theme, Time, TopBarControl, UniversalPasswordStorage,
    UnsaltedMd5PasswordStorage, WikingDefaultDataClass, WikingResolver,
    ajax_response, format_http_date, generate_random_string, log, module,
    parse_http_date, pdf_document, send_mail, serve_file, validate_email_address, breakpoint,
)

from .request import (  # noqa: F401
    ClosedConnection, FileUpload, Request, Role, Roles, ServerInterface, User,
)
from .modules import (  # noqa: F401
    ActionHandler, Documentation, Module, Reload, RequestHandler, Resources,
    Robots, Search, Session, SiteIcon, SubmenuRedirect,
)
from .db import (  # noqa: F401
    APIProvider, CachedTables, CachingPytisModule, CbCachingPytisModule, DBException,
    PytisModule, PytisRssModule, RssModule,
)

from .application import Application  # noqa: F401
from .export import Exporter, MinimalExporter  # noqa: F401
from .handler import Handler  # noqa: F401

from lcg import log as debug  # noqa: F401

# Initialize the global configuration object 'wiking.cfg'.
from .configuration import Configuration, ApplicationConfiguration  # noqa: F401
cfg = Configuration()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2012-2017 OUI Technology Ltd.
# Copyright (C) 2019-2020 Tomáš Cerha <t.cerha@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import copy
import os
import random
import re
import sys
import time
import unittest
import webtest

import wiking.wsgi_interface

import http.cookiejar
import urllib.parse


class _TestBase(unittest.TestCase):

    class _Visited(set):

        def __call__(self, url):
            if url in self:
                visited = True
            else:
                self.add(url)
                visited = False
            return visited

    class _RandomlyVisited:

        def __init__(self, transform=None):
            self._visited = dict()
            if transform is None:
                def transform(url):
                    return None
            self._transform = transform

        def __call__(self, url):
            transformed_url = self._transform(url)
            if transformed_url == url:
                visited = transformed_url in self._visited
                self._visited[transformed_url] = True
            else:
                n = self._visited.get(transformed_url, 0) + 1
                self._visited[transformed_url] = n
                if n == 1:
                    visited = False
                else:
                    visited = random.random() > 1.0 / n
            return visited

        def add(self, url):
            url = self._transform(url)
            self._visited[url] = self._visited.get(url, 0) + 1

        _numeric_suffix_regexp = re.compile('^(.*/)[0-9]+$')

        @classmethod
        def numeric_suffix_transformer(class_, url):
            match = class_._numeric_suffix_regexp.match(url)
            if match is not None:
                url = url[:match.end(1)] + '*'
            return url

    @classmethod
    def set_options(class_, config_file, host, options=None):
        class_._config_file = config_file
        class_._host = host
        class_._options = options

    def setUp(self):
        self._options = self._process_options(self.__class__._options)

    def _process_options(self, options):
        if options is None:
            options = object()
        for k, v in self._DEFAULT_OPTIONS:
            if getattr(options, k, None) is None:
                setattr(options, k, v)
        return options

    def _credentials(self, index=0):
        user_string = self._options.user
        if user_string is None:
            return None, None
        users = user_string.split(':')
        try:
            the_user = users[index]
        except IndexError:
            the_user = users[0]
        password_string = self._options.password
        if password_string is None:
            return the_user, None
        passwords = password_string.split(':')
        try:
            the_password = passwords[index]
        except IndexError:
            the_password = passwords[0]
        return the_user, the_password

    def _find_link(self, browser, description=None, index=None, verbose=False):
        if description is None:
            content_matcher = None
        else:
            content_matcher = re.compile(description).search
        links = []
        for element in self._find_all_links(browser):
            href = self._attribute(element, 'href')
            self._info("Element: %s" % (href,))
            if not href:
                self._info("  Skipped: no href")
                continue
            if href.startswith('#'):
                self._info("  Skipped: internal link")
                continue
            if href.startswith('javascript:'):
                self._info("  Skipped: JavaScript link")
                continue
            if href.startswith('mailto:'):
                self._info("  Skipped: mail link")
                continue
            if content_matcher is not None and content_matcher(self._element_text(element)) is None:
                self._info("  Skipped: doesn't match")
                continue
            self._info("  Accepted")
            url = urllib.parse.urljoin(self._current_url(browser), href)
            links.append(url)

        def exception_args():
            return (description, browser,) if self._options.verbose else ()
        if not links:
            raise IndexError("No matching link found", *exception_args())
        if index is None:
            if len(links) > 1:
                raise IndexError("Multiple matching links", *exception_args())
            index = 0
        if index is True:
            return links
        return links[index]

    def _filter_form(self, browser, options):
        form_fields = [o[0] for o in options]
        for field, text in options:
            form = self._find_form(browser, fields=form_fields)
            self._set_select_field(form, field, text=text)
            self._ajax_delay()
        form = self._find_form(browser, fields=form_fields)
        return self._submit_form(form, {'class': 'apply-filters'})

    def _click(self, browser, description=None, index=None, status=None, verbose=False,
               follow=False):
        url = self._find_link(browser, description=description, index=index, verbose=verbose)
        if follow:
            browser = self._get_follow(url)
        else:
            browser = self._get(url, status=status)
        return browser

    def _click_all(self, browser, visited=None, status=None, verbose=False, ignored=None):
        host = self._host
        all_responses = []
        for url in self._find_link(browser, index=True):
            hostname = urllib.parse.urlparse(url).hostname
            if hostname and hostname != host:
                continue
            if ignored is not None and ignored(url):
                continue
            if visited is None or not visited(url):
                all_responses.append(self._get(url, status=status))
        return all_responses

    def _info(self, message):
        if self._options.verbose:
            sys.stdout.write('%s\n' % (message,))

    def _warning(self, message):
        sys.stderr.write('WARNING: %s\n' % (message,))


class Test(_TestBase):

    def setUp(self):
        super(Test, self).setUp()
        self._headers = self._make_headers()
        self._environment = self._make_environment()
        self._cookies = http.cookiejar.CookieJar()
        self._application = webtest.TestApp(wiking.wsgi_interface.application,
                                            cookiejar=self._cookies)
        self._set_language()

    def _make_headers(self):
        headers = {'Host': self._host,
                   'User-Agent': 'Wiking Tester'}
        return headers

    def _make_environment(self):
        return {'wiking.config_file': self._config_file}

    def _set_language(self):
        language = self._options.language
        if language is not None:
            self._headers['Accept-Language'] = language

    def _default_request_kwargs(self):
        return dict(headers=self._headers, extra_environ=self._environment)

    def _current_url(self, response):
        return response.request.url

    def _get(self, path, status=None):
        if self._options.verbose:
            self._info('GET: %s' % (path,))
        return self._application.get(path, status=status, **self._default_request_kwargs())

    def _get_follow(self, path):
        response = self._get(path)
        return self._follow(response, path=path)

    def _follow(self, response, path=None):
        for i in range(10):
            if response.status_int != 302:
                return response
            response = response.follow(headers=self._headers)
        raise Exception("Too many redirections", path)

    def _find_all_links(self, response):
        return response.html.find_all('a')

    def _find_form(self, response, fields=(), check_found=True):
        for form_id in response.forms:
            form = response.forms[form_id]
            for f in fields:
                if isinstance(f, tuple):
                    name, value = f
                else:
                    name, value = f, None
                form_field = form.get(name, index=0, default=None)
                if form_field is None:
                    break
                if value is not None and form_field.value != value:
                    break
            else:
                return form
        self.assertFalse(check_found)
        return None

    def _find_elements(self, response, tag, attributes=None):
        if attributes is None:
            attributes = {}
        obj = response.html if isinstance(response, webtest.response.TestResponse) else response
        return obj.find_all(tag, **attributes)

    def _attribute(self, element, name):
        return element.get(name)

    def _element_text(self, element):
        return element.decode_contents()

    def _set_field(self, form, field, value):
        form[field] = value

    def _set_select_field(self, form, field, value=None, text=None):
        kwargs = dict(text=text) if text is not None else dict(value=value)
        form[field].select(**kwargs)

    def _submit_form(self, form, follow=True, **kwargs):
        form_kwargs = copy.copy(self._default_request_kwargs())
        form_kwargs.update(kwargs)
        response = form.submit(**form_kwargs)
        if follow and response:
            response = response.maybe_follow()
        return response

    def _contains(self, response, text):
        return text in response

    def _search(self, response, regexp):
        if isinstance(regexp, str):
            regexp = re.compile(regexp)
        return regexp.search(response.text)


class BrowserTest(_TestBase):

    def setUp(self):
        super(BrowserTest, self).setUp()
        profile = self._options.profile
        kwargs = {}
        if profile:
            profile_dir = os.path.expanduser('~/.mozilla/firefox')
            profile_candidates = [os.path.join(profile_dir, d) for d in os.listdir(profile_dir)
                                  if d.endswith('.' + profile)]
            if not profile_candidates:
                raise Exception("Profile %s not found" % (profile,))
            elif len(profile_candidates) > 1:
                raise Exception("Multiple directories matching profile %s:" % (profile,),
                                profile_candidates)
            kwargs['profile'] = profile_candidates[0]
        import splinter
        self._browser = splinter.Browser(**kwargs)
        self._set_language()
        self._ajax_delay_seconds = 1
        self._ajax_timeout_seconds = 10

    def tearDown(self):
        self._browser.quit()
        super(BrowserTest, self).tearDown()

    def _set_language(self):
        # Not possible in a general way
        pass

    def _current_url(self, browser):
        return browser.url

    def _get(self, path, status=None):
        browser = self._browser
        if path.startswith('/'):
            url = 'http://%s%s' % (self._host, path,)
        else:
            url = path
        browser.visit(url)
        status_code = browser.status_code
        if status is None:
            self.assertTrue(status_code.is_success())
        else:
            self.assertEqual(status_code, status)
        return browser

    def _get_follow(self, path):
        return self._get(path)

    def _find_elements(self, browser, tag, attributes=None):
        xpath = 'descendant::' + tag
        if attributes:
            equations = ['attribute::%s="%s"' % (k, v.replace('"', '\\"'),)
                         for k, v in attributes.items()]
            xpath += '[%s]' % (' and '.join(equations),)
        return browser.find_by_xpath(xpath)

    def _find_all_links(self, browser):
        return browser.find_link_by_partial_text('')

    def _find_form(self, browser, fields=(), check_found=True):
        for form in browser.find_by_tag('form'):
            if not form.visible:
                continue
            for f in fields:
                if isinstance(f, tuple):
                    name, value = f
                else:
                    name, value = f, None
                form_fields = form.find_by_name(name)
                if not form_fields:
                    break
                if value is not None and all([f.value != value for f in form_fields]):
                    break
            else:
                return form
        self.assertFalse(check_found)
        return None

    def _attribute(self, element, name):
        return element[name]

    def _element_text(self, element):
        return element.text

    def _set_field(self, form, field, value):
        form.find_by_name(field)[0].fill(value)

    def _set_select_field(self, form, field, value=None, text=None):
        select_field = form.find_by_name(field)[0]
        if text is not None:
            for element in select_field.find_by_xpath('descendant::option[@value]'):
                if element.text == text:
                    value = element['value']
                    break
            else:
                raise KeyError(field, text)
        select_field.select(value)

    def _ajax_delay(self, seconds=None, text=None):
        if text is None:
            time.sleep(seconds or self._ajax_delay_seconds)
        else:
            if seconds is None:
                seconds = self._ajax_timeout_seconds
            return self._browser.is_text_present(text, wait_time=seconds)

    def _attach_file(self, form, field, filename):
        form.attach_file(field, filename)

    def _click_element(self, element):
        element.click()

    def _submit_form(self, form, attributes=None, follow=None):
        if attributes is None:
            attributes = {}
        if 'type' not in attributes:
            attributes['type'] = 'submit'
        buttons = self._find_elements(form, 'button', attributes)
        if not buttons:
            raise Exception("No button found")
        buttons[0].click()
        return self._browser

    def _contains(self, browser, text):
        return browser.is_text_present(text)

    def _search(self, browser, regexp):
        if isinstance(regexp, str):
            regexp = re.compile(regexp)
        return regexp.search(browser.html)


def parse_options():
    parser = argparse.ArgumentParser()
    parser.add_argument('-l', '--language', dest='language', metavar='LANGUAGE',
                        help="try to use given LANGUAGE")
    parser.add_argument('--profile', dest='profile', metavar='PROFILE',
                        help="use given web browser PROFILE")
    parser.add_argument('-u', '--user', dest='user', metavar='USER[:USER...]',
                        help="use given USER(s) in login forms")
    parser.add_argument('-p', '--password', dest='password', metavar='PASSWORD[:PASSWORD...]',
                        help="use given PASSWORD(s) in login forms")
    parser.add_argument('-v', '--verbose', dest='verbose', action='store_true', default=False,
                        help="be verbose about some actions")
    parser.add_argument('config_file', metavar='CONFIG-FILE',
                        help="Wiking application configuration file")
    parser.add_argument('host', metavar='HOST',
                        help="HTTP host name")
    parser.add_argument('unittest_options', metavar='UNITTEST-OPTIONS', nargs='*')
    args = parser.parse_args()
    return args


def main():
    args = parse_options()
    _TestBase.set_options(args.config_file, args.host, options=args)
    argv = [sys.argv[0]] + args.unittest_options
    unittest.main(argv=argv)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2006-2017 OUI Technology Ltd.
# Copyright (C) 2019-2021, 2024 Tomáš Cerha <t.cerha@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Definition of default Wiking application and its API."""

import re
import os
import sys
import pytis
import wiking
import datetime
from xml.sax import saxutils

import lcg
from wiking import OPR, log

_ = lcg.TranslatableTextFactory('wiking')


class Application(wiking.Module):
    """Define Wiking application behavior.

    Wiking application is itself a Wiking module.  This module defines the
    basic application behavior.  You may customize your application by
    overriding this module and implementing the methods which form the
    Application API by your own means.

    """

    _MAPPING = {'_doc': 'Documentation',
                '_resources': 'Resources',
                'favicon.ico': 'SiteIcon',
                'robots.txt': 'Robots',
                }
    """Defines static assignment of modules responsible for handling distinct URI paths.

    The value is a dictionary, where keys are uri's and values are the names of
    the responsible modules as strings.  Only the first part of the request uri
    is used to determine the module.  Other parts (after the first slash) may
    be used by the module for further resolution of the request.

    This constant is used by the default implementation of
    'Application.handle()' and 'Application.module_uri()'.  See their
    documentation for more information.

    """

    _PREFERRED_LANGUAGE_COOKIE = 'wiking_preferred_language'

    def __init__(self, *args, **kwargs):
        super(Application, self).__init__(*args, **kwargs)
        self._mapping = dict(self._MAPPING)
        self._reverse_mapping = dict([(v, k) for k, v in self._MAPPING.items()])
        uri = self._reverse_mapping.get('Resources')
        if wiking.cfg.resources_version is not None and uri is not None:
            del self._mapping[uri]
            uri += '-' + wiking.cfg.resources_version
            self._mapping[uri] = 'Resources'
            self._reverse_mapping['Resources'] = uri

    def initialize(self, req):
        """Perform application specific initialization.

        This method is called once, just after Application instance creation.
        It is supposed to perform any application specific initialization

        Note, that 'req' is the first request that triggers initialization of
        the application, but the application will live much longer and serve a
        number of other requests which follow.  But this method is called only
        once for the first request.

        The default implementation does nothing.

        """
        pass

    def handle(self, req):
        """Handle the request.

        The Wiking Handler passes the request to the current application for
        further processing.  All errors and exceptions are handled by the
        handler.

        The return value and possible exceptions is the same as described in
        'RequestHandler.handle()'.

        The default implementation uses static mapping of request paths (URI)
        to wiking modules defined by the class constant '_MAPPING' to
        determine which module is responsible for processing the request and
        passes the request along to an instance of this module.  'NotFound' is
        raised if the mapping doesn't define an item for the current request
        URI.  You may re-implement this method if a different logic is more
        suitable for your application.

        """
        if not req.unresolved_path:
            # The request to the server's root URI is automatically redirected
            # to the first menu item's URI.  Redirection to hidden items is
            # avoided particularly because of default items added automatically
            # to the menu in CMS, such as _registration.  If the real menu is
            # empty, we don't want these items to be used for redirection.
            menu = [item for item in self.menu(req) if not item.hidden()]
            # The authentication must be performed before redirection, because
            # the current request may include login/logout commands which would
            # be discarded by the redirection.
            user = req.user()
            if menu:
                raise wiking.Redirect(menu[0].id())
            elif not user:
                # Here we handle the common case, that the menu is empty for
                # non-authenticated users and will be non-empty after the user
                # authenticates.  This presumption may not be theoretically
                # valid in all cases, but it typically works as expected.  If
                # the application doesn't want this behavior, it should not
                # return an empty menu.
                raise wiking.AuthenticationError()
            else:
                raise wiking.Forbidden()
        identifier = req.unresolved_path[0]
        try:
            modname = self._mapping[identifier]
        except KeyError:
            raise wiking.NotFound()
        mod = wiking.module(modname)
        assert isinstance(mod, wiking.RequestHandler)
        req.unresolved_path.pop(0)
        return req.forward(mod)

    def module_uri(self, req, modname):
        """Return the base URI of given Wiking module (relative to server root).

        The argument 'modname' is the Wiking module name as a string.

        If the module has no definite global path within the application, None
        may be returned.  The exact behavior depends on the particular
        application in use.  Please see the documentation of this method in the
        relevant application class (such as
        'wiking.cms.Application.module_uri()' for applications built on Wiking
        CMS) for more details.

        The default implementation performs a reverse lookup in the '_MAPPING'
        dictionary.  None is returned when there is no mapping item for the
        module, or when there is more than one item for the same module (which
        is also legal).

        This method should not be called directly by application code.  Use
        'wiking.Request.module_uri()' instead (which calls this method
        internally if necessary).

        """
        identitier = self._reverse_mapping.get(modname)
        return identitier and '/' + identitier or None

    def authenticate(self, req, login, password, auth_type):
        """Return 'User' instance if given authentication credentials are valid.

        Arguments:
          req -- current request object
          login -- user's login name as a string
          password -- user supplied password as a string
          auth_type -- Short string describing the authentication method.  Each
            'wiking.AuthenticationProvider' subclass uses its specific string,
            such as "Cookie" for 'wiking.CookieAuthenticationProvider' etc.
            Only informational (for logging etc.).

        Returns a 'User' instance corresponding to given login name and
        password if the password is valid for given user or None if not.

        This method must be implemented by derived applications.  The default
        implementation always returns None.

        """
        return None

    def login_hook(self, req, user):
        """Hook executed after succesfull authentication.

        Arguments:

          req -- current request object
          user -- 'User' instance of the authenticated user

        The default implementation does nothing.  Derived applications may
        override it.

        """
        pass

    def logout_hook(self, req, user):
        """Hook executed when session is closed explicitly by user.

        Arguments:

          req -- current request object
          user -- 'User' instance of the authenticated user

        The default implementation does nothing.  Derived applications may
        override it.

        """
        pass

    def contained_roles(self, role):
        """Return the sequence of user roles contained in given role.

        Arguments:

          role -- User role as a 'Role' instance.

        In general, user roles may be contained in each other.  This means that
        user's membersip in one role (let's say role A) may automatically imply
        his membership in other roles (B and C for example).  Roles B and C
        contain role A in this example.  If role containment is supported by
        the application, this method must be implemented and must return the
        list of all contained roles (including the given role itself and also
        transitively contained roles).

        In any case there may be no cycles in role containment, i.e. no role
        may contain itself, including transitive relations.  For instance, role
        A may not contain A; or if A contains B and B contains C then C may not
        contain A nor B nor C.

        """
        return ()

    def site_title(self, req):
        """Return site title as a string.

        This method returns the value of configuration option 'site_title' by
        default.  It may be overriden to change the title dynamically.

        """
        return wiking.cfg.site_title

    def site_subtitle(self, req):
        """Return site subtitle as a string or None.

        Site subtitle is normally appended to site title separated by a dash
        (if not 'None').  It is, however, exported in a differnt html element
        to allow independent styling.

        This method returns the value of configuration option 'site_subtitle'
        by default.  It may be overriden to change the subtitle dynamically.

        """
        return wiking.cfg.site_subtitle

    def top_content(self, req):
        """Return the content displayed on the left side of the top bar.

        The returned value is a list of 'lcg.Content' instances or any other
        value acceptable by lcg.coerce()

        The content must behave nicely with the controls displayed on the right
        side of the top bar (see 'top_controls()').  This particularly means to
        to style the content to be responsive to the actual display size
        through CSS.  Otherwise the top bar may not fit to the visible area
        (hiding the controls at the right edge).

        The default implementation returns a list with just one element -- the
        'lcg.HtmlContent' instance displaying the site title and subtitle as
        returned by the methods 'site_title()' and 'site_subtitle()'.

        """
        def site_title(context, element, title, subtitle):
            g = context.generator()
            content = g.strong(title, cls='title')
            if subtitle:
                content += (g.strong(g.noescape(' &ndash; '), cls='separator') +
                            g.strong(subtitle, cls='subtitle'))
            return g.div(g.a(content, href='/'), id='site-title')
        return [lcg.HtmlContent(site_title, self.site_title(req), self.site_subtitle(req))]

    def top_controls(self, req):
        """Return the controls displayed on the right side of the top bar.

        The top bar displays the content returned by 'top_content()' (which
        normally is site title and subtitle) on the left and controls on the
        right.  Controls are simple (usually interactive) widgets for
        performing site global actions, such as logging in/out or switching
        languages (these two controls are returned by default).  Derived
        applications may override this method to add custom controls, customize
        the built-in controls or re-arrange the order of controls on the top
        bar.

        Make sure the controls don't take too much space or make them
        responsive to the actual display size through CSS.

        Any content acceptable by 'lcg.coerce()' may be returned ('lcg.Content'
        instance, str, or their sequence).  Consider using the base
        class 'wiking.TopBarControl' for definition of custom controls.

        The default implementation returns a list of two instances:
        'wiking.LoginControl' and 'wiking.LanguageSelection'.

        """
        return [wiking.LoginControl(), wiking.LanguageSelection(), wiking.MaximizedModeControl()]

    def menu(self, req):
        """Return the main navigation menu hierarchy.

        Arguments:

          req -- the current request object.

        Returns a sequence of 'MenuItem' instances representing the main menu
        hierarchy.

        The menu structure should usually remains unchanged throughout the
        application or at least throughout its major states, but this is just a
        common practice, not a requirement.  The application may decide to
        return a different menu for each request.

        """
        return ()

    def panels(self, req, lang):
        """Return a list of 'Panel' instances representing panels displayed on the page.

        The set of panels will usually be constant throughout the application,
        but this is just a common practice, not a requirement.  The application
        may decide to return a different set of panels for each request.

        See the Navigation section of the Wiking User's documentation for
        general information about panels.

        """
        return []

    def languages(self):
        """Return a list of all languages supported by the application.

        Returns a list of all supported languages as the corresponding alpha-2
        language codes.

        You should check, whether the translations for all the specified
        languages are available in the gettext catalog for Wiking and also for
        all the components you are using, such as LCG and Pytis.

        """
        return ['en']

    def preferred_languages(self, req):
        """Return a list of languages accepted by the user in the order of their preference.

        The list of prefered languages allows wiking applications to serve
        multilingual content using the automatic Content Negotiation techique
        with a possibility for the user to switch between the available
        language variants manually.  You should never call this method directly
        in a Wiking application.  You rather want to call
        'Request.preferred_languages()'.

        You may need to override this method to change its logic completely or
        partially for given application if the default logic described below
        doesn't suit your needs.

        The default implementation uses a combination of user preferences set
        in 'Accept-Language' HTTP headers, explicit language switching and
        global defaults set in configuration.

        If the application supports language switching (the small widget at the
        top right corner in the default layout), the language selected there
        has the highest precedence (the selected value is saved in a cookie).
        The languages set through the HTTP header Accept-Language in browsers
        preferences immediately follow in the order of their precedence.  The
        last option (lowest precedence) is the default language configured for
        the server through 'default_language' and 'default_language_by_domain'
        configuration options.

        """
        if req.has_param('setlang'):
            # Use the language selected explicitly by the user in this request.
            selected = str(req.param('setlang'))
            req.set_param('setlang', None)
            req.set_cookie(self._PREFERRED_LANGUAGE_COOKIE, selected)
        elif req.has_param('fb_locale'):
            # This parameter is used by Facebook's link sharing robot, which
            # explores the linked page when it is shared to determine what to
            # show on the page.
            selected = str(req.param('fb_locale').split('_')[0])
        else:
            # Use the language selected previously (if at all).
            selected = req.cookie(self._PREFERRED_LANGUAGE_COOKIE)
            if selected:
                selected = str(selected)
        if selected:
            languages = [selected]
        else:
            languages = []
        languages.extend([lang for lang in req.accepted_languages() if lang != selected])
        default = wiking.cfg.default_language_by_domain.get(req.server_hostname(),
                                                            wiking.cfg.default_language)
        if default and default not in languages:
            languages.append(default)
        return languages

    def stylesheets(self, req):
        """Return the list of all available style sheets as 'lcg.Stylesheet' instances.

        The application is responsible for handling the returned stylesheets
        (serving them to clients).  Wiking provides a generic 'Resources'
        module for this purpose.

        The default implementation returns the 'default.css' stylesheet defined
        by Wiking.

        """
        return [lcg.Stylesheet('default.css')]

    def registration_uri(self, req):
        """Return the URI for new user registration or None if registration is not allowed."""
        return None

    def forgotten_password_uri(self, req):
        """Return the forgotten password link URI or None if password reset is not implemented."""
        return None

    def password_change_uri(self, req):
        """Return the change password URI or None if the link should not appear in login control."""
        return None

    def login_is_email(self, req):
        """Return True iff e-mail addresses are used as login names.

        Using e-mail addresses as login names is quite typical.  If this is
        indicated by returning True from this method, the user interface will
        use more appropriate labels for login controls to avoid user's
        confusion.

        """
        return False

    def body_class_names(self, req):
        """Return a sequence of CSS class names to assign to the BODY element.

        This allows application specific class names to be assigned to the top
        level BODY element in the HTML export of all pages.

        """
        return ()

    def login_dialog_top_content(self, req):
        """Return the content displayed above the login dialog as 'lcg.Content' element(s).

        Any content acceptable by 'lcg.coerce()' may be returned.

        """
        return None

    def login_dialog_bottom_content(self, req):
        """Return the content displayed below the login dialog as 'lcg.Content' element(s).

        Any content acceptable by 'lcg.coerce()' may be returned.

        """
        return self.login_dialog_content(req)

    def login_dialog_content(self, req):
        """Deprecated.  Use 'login_dialog_bottom_content()'."""
        return None

    def _powered_by_wiking(self, req):
        import wiking
        # Translators: Website idiom. This is followed by information on the underlying software
        # tools.  Means: This website runs on [...this and that software...].
        return (_("Powered by"), ' ',
                lcg.link('http://www.freebsoft.org/wiking', 'Wiking'), ' ', wiking.__version__)

    def _accessibility_statement_link(self, req):
        """Return lcg.Content containing a link to the Wiking Accessibility Statement or None.

        None is returned when the Documentation module (which normally serves
        the accessibility statement text) is not available (not mapped to a
        valid URI by the application).

        """
        doc_uri = req.module_uri('Documentation')
        if doc_uri:
            uri = doc_uri + '/wiking/user/accessibility'
            return lcg.HtmlContent(lambda context, element: context.generator().a(
                # accesskey not supported by generic lcg links.
                _("Accessibility Statement"), href=uri, accesskey='0',
            ))
        else:
            return None

    def menu_panel_title(self, req):
        """Return the displayed title of the left side hierarchical menu as a string."""
        # Translators: Heading of webpage left side hierarchical navigation
        # menu containing a list of links to pages in this web section
        return _("In this section:")

    def menu_panel_tooltip(self, req):
        """Return the tooltip of the left side hierarchical menu as a string."""
        return _("Local navigation")

    def menu_panel_bottom_content(self, req):
        """Return the additional content to be displayed in the hierarchical menu panel.

        Any content acceptable by 'lcg.coerce()' may be returned.  The content
        will be placed under the actual hierarchical menu list (at the bottom
        of the panel).

        """
        return None

    def right_panels_bottom_content(self, req):
        """Return the additional content to be displayed under right side panels.

        Any content acceptable by 'lcg.coerce()' may be returned.  The content
        will be placed under the list of panels on the right side (if panels
        are not hidden).

        """
        return None

    def bottom_bar_left_content(self, req):
        """Return the content displayed on the left side of the bottom bar above the page footer.

        Any content acceptable by 'lcg.coerce()' may be returned.

        """
        return self._powered_by_wiking(req)

    def bottom_bar_right_content(self, req):
        """Return the content displayed on the right side of the bottom bar above the page footer.

        Any content acceptable by 'lcg.coerce()' may be returned.

        """
        return lcg.join((lcg.link('/privacy-policy', _("Privacy policy")),
                         self._accessibility_statement_link(req)), separator=' \u2022 ')

    def footer_content(self, req):
        """Return the content displayed in page footer as 'lcg.Content' element(s).

        Any content acceptable by 'lcg.coerce()' may be returned.

        """
        return lcg.p(_("Contact:"), ' ',
                     lcg.link("mailto:" + wiking.cfg.webmaster_address,
                              wiking.cfg.webmaster_address))

    def _send_bug_report(self, req, error, info, address):
        """Send traceback of given InternalServerError to given e-mail address."""
        def format_param(param):
            if param in ('passwd', 'password', 'initial_password'):
                value = '<password hidden>'
            else:
                try:
                    value = req.param(param)
                except UnicodeDecodeError as e:
                    # TODO: WsgiRequest.param() currently fails on invalid characters
                    # Once the problem is handled there, this may not be necessary.
                    value = "<UnicodeDecodeError: %s>" % e
                else:
                    if isinstance(value, wiking.FileUpload):
                        value = '%s; mime_type="%s"' % (value.filename(), value.mime_type())
                    elif isinstance(value, tuple):
                        value = repr(value)
                    else:
                        lines = value.splitlines()
                        if len(lines) > 1:
                            value = lines[0][:40] + '... (trimmed; total %d lines)' % len(lines)
                    if len(value) > 40:
                        value = value[:40] + '... (trimmed; total %d chars)' % len(value)
            return "   %s = %s" % (saxutils.escape(param), saxutils.escape(value))
        def maybe_link(value):
            if value and (value.startswith('http://') or value.startswith('https://')):
                value = '<a href="%s">%s</a>' % (value, value)
            return value
        header = (
            ("URI", info['abs_uri']),
            ("HTTP referer", info['referer']),
            ('Server software', info['server_software']),
            ('HTTP method', info['method']),
            ('Reference ID', info['ref_id']),
            ("Request parameters", "\n" + "\n".join(map(format_param, req.params()))),
        )
        text = "\n\n".join((
            "\n".join(["%s: %s" % pair for pair in header]),
            error.traceback(detailed=True, format='text'),
        ))
        html = "<html><pre>%s</pre>%s</html>" % (
            "\n".join(["%s: %s" % (label, maybe_link(value)) for label, value in header]),
            error.traceback(detailed=True, format='html'),
        )
        return wiking.send_mail(address, subject='Wiking Error: ' + error.signature(),
                                text=text, html=html,
                                headers=(('Reply-To', address),
                                         ('X-Wiking-Bug-Report-From',
                                          wiking.cfg.server_hostname)))

    def report_error(self, req, error):
        """Invoked for all errors to record their occurance for later review.

        This method is invoked for every 'ReqestError' exception raised within
        the application in order to record it and report the problem to the
        operator/maintainer/developer of the application.

        All errors are written into server's error log.  The amount of
        information and their format is controlled by the 'log_format'
        configuration option.

        Additionally, if the configuration option 'bug_report_address' is set,
        detailed cgitb traceback of the error is sent to the configured email
        address.

        Personal information, such as user's login, IP address or User Agent
        are intentionally not included in the e-mail but may be logged (if
        'log_format' contains them) and paired with the e-mailed traceback
        using the 'ref_id' value.

        For 'InternalServerError' a brief traceback (without the details from
        cgitb) is also written to the error log.  This allows jumping right to
        the code when the error log is observerd in an IDE during development.
        If 'debug' configuration option is set, the traceback is also logged
        for other error types for the same purpose.

        """
        if not isinstance(error, wiking.AuthenticationRedirect):
            info = dict(
                server_hostname=req.server_hostname(),
                uri=req.uri(),
                abs_uri=req.server_uri(current=True) + req.uri(),
                user=(req.user() and req.user().login() or 'anonymous'),
                remote_host=req.remote_host(),
                referer=req.header('Referer'),
                user_agent=req.header('User-Agent'),
                method=req.method(),
                server_software=('Wiking %s, LCG %s, Pytis %s' %
                                 (wiking.__version__, lcg.__version__, pytis.__version__)),
                error_type=error.__class__.__name__,
                ref_id=wiking.generate_random_string(10),
            )
            log(OPR, wiking.cfg.log_format % info)
            if isinstance(error, wiking.InternalServerError):
                log(OPR, error.traceback())
                address = wiking.cfg.bug_report_address
                if address:
                    err = self._send_bug_report(req, error, info, address)
                    if err:
                        log(OPR, "Failed sending error details to %s:" % address, err)
                    else:
                        log(OPR, "Error details sent to %s." % address)
            elif wiking.cfg.debug:
                log(OPR, error.traceback())
